# -*- coding: utf-8 -*-
"""
Project Points CLI
"""
import click
import logging

from reV.config.project_points import ProjectPoints
from reV.utilities.exceptions import ProjectPointsValueError
from reV.utilities import log_versions
from reV import __version__

from rex.utilities.cli_dtypes import STR
from rex.utilities.loggers import init_logger
from rex.utilities.utilities import (dict_str_load, safe_json_load)

logger = logging.getLogger(__name__)


def _parse_lat_lons(lat_lon_fpath, lat_lon_coords):
    """
    Parse CLI latitude and longitude inputs for ProjectPoints

    Parameters
    ----------
    lat_lon_fpath : str
        File path to .csv or .json containing latitude, longitude coordinates
        of interest
    lat_lon_coords : tuple
        (lat, lon) coordinates of interest

    Returns
    -------
    lat_lons : str | tuple
        File path to .csv or .json containing latitude, longitude coordinates
        of single set of coordinates to convert to ProjectPoints
    """
    lat_lons = None
    msg = None
    if lat_lon_fpath is not None:
        lat_lons = lat_lon_fpath

    if lat_lon_coords:
        if lat_lons is None:
            lat_lons = lat_lon_coords
        else:
            msg = ("Both a 'lat-lon-fpath' and a pair of 'lat-lon-coords' "
                   "were supplied, but only one can be used to create "
                   "ProjectPoints!")

    if lat_lons is None:
        msg = ("A 'lat-lon-fpath' or a pair of 'lat-lon-coords' must be "
               "supplied in order to create ProjectPoints!")

    if msg is not None:
        logger.error(msg)
        raise ProjectPointsValueError(msg)

    return lat_lons


def _parse_regions(regions, region, region_col):
    """
    Parse CLI regions inputs for ProjectPoints

    Parameters
    ----------
    regions : str
        json string or file path to .json containing regions of
        interest in the form {'region': 'region_column'}
    region : str
        Region to extract
    region_col : str
        Meta column to search for region

    Returns
    -------
    regions : dict
        Dictionary of region_col: regions to generate project points
    """
    msg = None
    if regions is not None:
        if regions.endwtih('.json'):
            regions = safe_json_load(regions)
        else:
            regions = dict_str_load(regions)

    if region is not None:
        if regions is None:
            regions = {region: region_col}
        else:
            if region in regions:
                msg = ("Multiple values for {}: {} were provided!"
                       .format(region, region_col))
            else:
                regions.update({region: region_col})

    if regions is None:
        msg = ("At least a single 'region' and 'region-col' must be "
               "supplied in order to create ProjectPoints!")

    if msg is not None:
        logger.error(msg)
        raise ProjectPointsValueError(msg)

    return regions


@click.group()
@click.version_option(version=__version__)
@click.option('--fpath', '-f', type=click.Path(), required=True,
              help='.csv file path to save project points to (required)')
@click.option('--res_file', '-rf', required=True,
              help=('Filepath to single resource file, multi-h5 directory, '
                    'or /h5_dir/prefix*suffix (required)'))
@click.option('--sam_file', '-sf', required=True,
              type=click.Path(exists=True), help='SAM config file (required)')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, fpath, res_file, sam_file, verbose):
    """reV ProjectPoints generator"""
    ctx.ensure_object(dict)
    ctx.obj['FPATH'] = fpath
    ctx.obj['RES_FILE'] = res_file
    ctx.obj['SAM_FILE'] = sam_file
    ctx.obj['VERBOSE'] = verbose

    if verbose:
        log_level = 'DEBUG'
    else:
        log_level = 'INFO'

    init_logger('reV.config.project_points', log_level=log_level)
    log_versions(logger)


@main.command()
@click.option('--lat_lon_fpath', '-llf', type=click.Path(exists=True),
              default=None,
              help=('File path to .csv or .json containing latitude, '
                    'longitude coordinates of interest'))
@click.option('--lat_lon_coords', '--llc', nargs=2, type=float, default=None,
              help='(lat, lon) coordinates of interest')
@click.pass_context
def from_lat_lons(ctx, lat_lon_fpath, lat_lon_coords):
    """Convert latitude and longitude coordinates to ProjectPoints"""
    lat_lons = _parse_lat_lons(lat_lon_fpath, lat_lon_coords)
    logger.info('Creating ProjectPoints from {} and saving to {}'
                .format(lat_lons, ctx.obj['FPATH']))
    pp = ProjectPoints.lat_lon_coords(lat_lons, ctx.obj['RES_FILE'],
                                      ctx.obj['SAM_FILE'])
    pp.df.to_csv(ctx.obj['FPATH'])


@main.command()
@click.option('--regions', '-regs', type=STR, default=None,
              help=('json string or file path to .json containing regions of '
                    'interest containing regions of interest'))
@click.option('--region', '-r', type=STR, default=None,
              help='Region to extract')
@click.option('--region_col', '-col', type=STR, default='state',
              help='Meta column to search for region')
@click.pass_context
def from_regions(ctx, regions, region, region_col):
    """Extract ProjectPoints for given geographic regions"""
    regions = _parse_regions(regions, region, region_col)
    logger.info('Creating ProjectPoints from {} and saving to {}'
                .format(regions, ctx.obj['FPATH']))
    pp = ProjectPoints.regions(regions, ctx.obj['RES_FILE'],
                               ctx.obj['SAM_FILE'])
    pp.df.to_csv(ctx.obj['FPATH'])


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running reV ProjecPoints CLI')
        raise



# -*- coding: utf-8 -*-
# pylint: disable=all
# flake8: noqa
"""
reV Supply Curve Aggregation command line interface (cli).
"""
import os
import click
import logging
import pprint
import time

from reV.config.supply_curve_configs import SupplyCurveAggregationConfig
from reV.pipeline.status import Status
from reV.supply_curve.tech_mapping import TechMapping
from reV.supply_curve.aggregation import Aggregation
from reV.supply_curve.sc_aggregation import SupplyCurveAggregation
from reV.utilities import ModuleName
from reV import __version__

from rex.multi_file_resource import MultiFileResource
from rex.utilities.hpc import SLURM
from rex.utilities.cli_dtypes import (STR, INT, FLOAT, STRLIST, FLOATLIST,
                                      STRFLOAT, STR_OR_LIST)
from rex.utilities.loggers import init_mult
from rex.utilities.utilities import dict_str_load, get_class_properties

logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('--name', '-n', default=os.path.basename(os.getcwd()),
              type=STR, show_default=True,
              help='reV Supply Curve Aggregation job name.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, name, verbose):
    """reV Supply Curve Aggregation Command Line Interface"""
    ctx.ensure_object(dict)
    ctx.obj['NAME'] = name
    ctx.obj['VERBOSE'] = verbose


@main.command()
def valid_config_keys():
    """
    Echo the valid SupplyCurveAggregation config keys
    """
    click.echo(', '.join(get_class_properties(SupplyCurveAggregationConfig)))


@main.command()
@click.option('--config_file', '-c', required=True,
              type=click.Path(exists=True),
              help='reV aggregation configuration json file.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def from_config(ctx, config_file, verbose):
    """Run reV SC aggregation from a config file."""

    # Instantiate the config object
    config = SupplyCurveAggregationConfig(config_file)

    # take name from config
    name = ctx.obj['NAME'] = config.name

    # Enforce verbosity if logging level is specified in the config
    if config.log_level == logging.DEBUG:
        verbose = True

    # initialize loggers
    init_mult(name, config.log_directory, modules=[__name__, 'reV', 'rex'],
              verbose=verbose)

    # Initial log statements
    logger.info('Running reV supply curve aggregation from config '
                'file: "{}"'.format(config_file))
    logger.info('Target output directory: "{}"'.format(config.dirout))
    logger.info('Target logging directory: "{}"'.format(config.log_directory))
    logger.debug('The full configuration input is as follows:\n{}'
                 .format(pprint.pformat(config, indent=4)))

    if config.execution_control.option == 'local':
        status = Status.retrieve_job_status(
            config.dirout,
            module=ModuleName.SUPPLY_CURVE_AGGREGATION,
            job_name=name
        )
        if status != 'successful':
            Status.add_job(
                config.dirout, module=ModuleName.SUPPLY_CURVE_AGGREGATION,
                job_name=name, replace=True,
                job_attrs={'hardware': 'local',
                           'fout': '{}.csv'.format(name),
                           'dirout': config.dirout})
            ctx.invoke(
                direct,
                excl_fpath=config.excl_fpath,
                gen_fpath=config.gen_fpath,
                econ_fpath=config.econ_fpath,
                res_fpath=config.res_fpath,
                tm_dset=config.tm_dset,
                excl_dict=config.excl_dict,
                res_class_dset=config.res_class_dset,
                res_class_bins=config.res_class_bins,
                cf_dset=config.cf_dset,
                lcoe_dset=config.lcoe_dset,
                h5_dsets=config.h5_dsets,
                data_layers=config.data_layers,
                resolution=config.resolution,
                excl_area=config.excl_area,
                power_density=config.power_density,
                area_filter_kernel=config.area_filter_kernel,
                min_area=config.min_area,
                friction_fpath=config.friction_fpath,
                friction_dset=config.friction_dset,
                cap_cost_scale=config.cap_cost_scale,
                out_dir=config.dirout,
                max_workers=config.execution_control.max_workers,
                sites_per_worker=config.execution_control.sites_per_worker,
                log_dir=config.log_directory,
                recalc_lcoe=config.recalc_lcoe,
                pre_extract_inclusions=config.pre_extract_inclusions,
                verbose=verbose)

    elif config.execution_control.option in ('eagle', 'slurm'):
        spw = config.execution_control.sites_per_worker
        ctx.obj['NAME'] = name
        ctx.obj['EXCL_FPATH'] = config.excl_fpath
        ctx.obj['GEN_FPATH'] = config.gen_fpath
        ctx.obj['ECON_FPATH'] = config.econ_fpath
        ctx.obj['RES_FPATH'] = config.res_fpath
        ctx.obj['TM_DSET'] = config.tm_dset
        ctx.obj['EXCL_DICT'] = config.excl_dict
        ctx.obj['RES_CLASS_DSET'] = config.res_class_dset
        ctx.obj['RES_CLASS_BINS'] = config.res_class_bins
        ctx.obj['CF_DSET'] = config.cf_dset
        ctx.obj['LCOE_DSET'] = config.lcoe_dset
        ctx.obj['H5_DSETS'] = config.h5_dsets
        ctx.obj['DATA_LAYERS'] = config.data_layers
        ctx.obj['RESOLUTION'] = config.resolution
        ctx.obj['EXCL_AREA'] = config.excl_area
        ctx.obj['POWER_DENSITY'] = config.power_density
        ctx.obj['AREA_FILTER_KERNEL'] = config.area_filter_kernel
        ctx.obj['MIN_AREA'] = config.min_area
        ctx.obj['FRICTION_FPATH'] = config.friction_fpath
        ctx.obj['FRICTION_DSET'] = config.friction_dset
        ctx.obj['CAP_COST_SCALE'] = config.cap_cost_scale
        ctx.obj['OUT_DIR'] = config.dirout
        ctx.obj['MAX_WORKERS'] = config.execution_control.max_workers
        ctx.obj['SITES_PER_WORKER'] = spw
        ctx.obj['LOG_DIR'] = config.log_directory
        ctx.obj['RECALC_LCOE'] = config.recalc_lcoe
        ctx.obj['PRE_EXTRACT_INCLUSIONS'] = config.pre_extract_inclusions
        ctx.obj['VERBOSE'] = verbose

        ctx.invoke(slurm,
                   alloc=config.execution_control.allocation,
                   memory=config.execution_control.memory,
                   feature=config.execution_control.feature,
                   walltime=config.execution_control.walltime,
                   conda_env=config.execution_control.conda_env,
                   module=config.execution_control.module,
                   sh_script=config.execution_control.sh_script)


@main.group(invoke_without_command=True)
@click.option('--excl_fpath', '-exf', type=STR_OR_LIST, required=True,
              help='Single exclusions file (.h5) or a '
              'list of exclusion files (.h5, .h5).')
@click.option('--gen_fpath', '-gf', type=STR, default=None,
              help='reV generation/econ output file. If not input, a generic '
              'exclusions aggregation will be run based on the res_fpath '
              'input.')
@click.option('--tm_dset', '-tm', type=STR, required=True,
              help='Dataset in the exclusions file that maps the exclusions '
              'to the resource data file being analyzed.')
@click.option('--econ_fpath', '-ef', type=STR, default=None,
              show_default=True,
              help='reV econ output file (optional argument that can be '
              'included if reV gen and econ data are being used from '
              'different files.')
@click.option('--res_fpath', '-rf', type=STR, default=None,
              show_default=True,
              help='Resource file (e.g. WTK or NSRDB), required if techmap '
              'dset is to be created or if gen_fpath is not input.')
@click.option('--excl_dict', '-exd', type=STR, default=None,
              show_default=True,
              help='String representation of a dictionary of exclusion '
              'LayerMask arguments {layer: {kwarg: value}} where layer '
              'is a dataset in excl_fpath and kwarg can be '
              '"exclude_values", "exclude_range", "include_values", '
              '"include_range", "include_weights", "force_include_values", '
              '"force_include_range", "use_as_weights", "weight", '
              '"exclude_nodata", and/or "nodata_value".')
@click.option('--res_class_dset', '-cd', type=STR, default=None,
              show_default=True,
              help='Dataset to determine the resource class '
              '(must be in gen_fpath).')
@click.option('--res_class_bins', '-cb', type=FLOATLIST, default=None,
              show_default=True,
              help='List of resource class bin edges.')
@click.option('--cf_dset', '-cf', type=STR, default='cf_mean-means',
              show_default=True,
              help='Dataset containing capacity factor values to aggregate.')
@click.option('--lcoe_dset', '-lc', type=STR, default='lcoe_fcr-means',
              show_default=True,
              help='Dataset containing lcoe values to aggregate.')
@click.option('--h5_dsets', '-hd', type=STRLIST, default=None,
              show_default=True,
              help='Additional datasets from the source gen/econ h5 files to '
              'aggregate.')
@click.option('--data_layers', '-d', type=STR, default=None,
              show_default=True,
              help='String representation of a dictionary of additional data '
              'layers to include in the aggregation e.g. '
              '{"slope": {"dset": "srtm_slope", "method": "mean"}}')
@click.option('--resolution', '-r', type=INT, default=64,
              show_default=True,
              help='Number of exclusion points along a squares edge to '
              'include in an aggregated supply curve point.')
@click.option('--excl_area', '-ea', type=FLOAT, default=None,
              show_default=True,
              help='Area of an exclusion pixel in km2. None will try to '
              'infer the area from the profile transform attribute in '
              'excl_fpath.')
@click.option('--power_density', '-pd', type=STRFLOAT, default=None,
              show_default=True,
              help='Power density in MW/km2 or filepath to variable power '
              'density csv file. None will attempt to infer a constant '
              'power density from the generation meta data technology. '
              'Variable power density csvs must have "gid" and '
              '"power_density" columns where gid is the resource gid '
              '(typically wtk or nsrdb gid). and the power_density column '
              'is in MW/km2.')
@click.option('--area_filter_kernel', '-afk', type=STR, default='queen',
              show_default=True,
              help='Contiguous area filter kernel name ("queen", "rook").')
@click.option('--min_area', '-ma', type=FLOAT, default=None,
              show_default=True,
              help='Contiguous area filter minimum area, default is None '
              '(No minimum area filter).')
@click.option('--friction_fpath', '-ff', type=STR, default=None,
              show_default=True,
              help='Optional h5 filepath to friction surface data. '
              'Must match the exclusion shape/resolution and be '
              'paired with the --friction_dset input arg. The friction data '
              'creates a "mean_lcoe_friction" output which is the nominal '
              'LCOE multiplied by the friction data.')
@click.option('--friction_dset', '-fd', type=STR, default=None,
              show_default=True,
              help='Optional friction surface dataset in friction_fpath.')
@click.option('--cap_cost_scale', '-cs', type=STR, default=None,
              show_default=True,
              help='Optional LCOE scaling equation to implement "economies '
              'of scale". Equations must be in python string format and '
              'return a scalar value to multiply the capital cost by. '
              'Independent variables in the equation should match the names '
              'of the columns in the reV supply curve aggregation table. '
              'This equation may use numpy functions with the package prefix '
              '"np". This will not affect offshore wind LCOE.')
@click.option('--out_dir', '-o', type=STR, default='./',
              show_default=True,
              help='Directory to save aggregation summary output.')
@click.option('--max_workers', '-mw', type=INT, default=None,
              show_default=True,
              help=("Number of cores to run summary on. None is all "
                    "available cpus"))
@click.option('--sites_per_worker', '-spw', type=INT, default=100,
              show_default=True,
              help="Number of sc_points to summarize on each worker")
@click.option('--log_dir', '-ld', type=STR, default='./logs/',
              show_default=True,
              help='Directory to save aggregation logs.')
@click.option('-recalc', '--recalc_lcoe', type=bool, default=True,
              show_default=True,
              help='Flag to recalculate LCOE from the multi-year mean '
              'capacity factor and annual energy production data. The recalc '
              '(which happens by default because it represents a more "True" '
              'LCOE) requires several datasets to be aggregated in the '
              'h5_dsets input: system_capacity, fixed_charge_rate, '
              'capital_cost, fixed_operating_cost, and '
              'variable_operating_cost.')
@click.option('-pre', '--pre_extract_inclusions', is_flag=True,
              help='Optional flag to pre-extract/compute the inclusion mask '
              'from the provided excl_dict, by default False. Typically '
              'faster to compute the inclusion mask on the fly with parallel '
              'workers.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def direct(ctx, excl_fpath, gen_fpath, tm_dset, econ_fpath, res_fpath,
           excl_dict, res_class_dset, res_class_bins,
           cf_dset, lcoe_dset, h5_dsets, data_layers, resolution, excl_area,
           power_density, area_filter_kernel, min_area, friction_fpath,
           friction_dset, cap_cost_scale, out_dir, max_workers,
           sites_per_worker, log_dir, recalc_lcoe,
           pre_extract_inclusions, verbose):
    """reV Supply Curve Aggregation Summary CLI."""
    sites_per_worker = sites_per_worker if sites_per_worker else 100

    name = ctx.obj['NAME']
    ctx.obj['EXCL_FPATH'] = excl_fpath
    ctx.obj['GEN_FPATH'] = gen_fpath
    ctx.obj['ECON_FPATH'] = econ_fpath
    ctx.obj['RES_FPATH'] = res_fpath
    ctx.obj['TM_DSET'] = tm_dset
    ctx.obj['EXCL_DICT'] = excl_dict
    ctx.obj['RES_CLASS_DSET'] = res_class_dset
    ctx.obj['RES_CLASS_BINS'] = res_class_bins
    ctx.obj['CF_DSET'] = cf_dset
    ctx.obj['LCOE_DSET'] = lcoe_dset
    ctx.obj['H5_DSETS'] = h5_dsets
    ctx.obj['DATA_LAYERS'] = data_layers
    ctx.obj['RESOLUTION'] = resolution
    ctx.obj['EXCL_AREA'] = excl_area
    ctx.obj['POWER_DENSITY'] = power_density
    ctx.obj['AREA_FILTER_KERNEL'] = area_filter_kernel
    ctx.obj['MIN_AREA'] = min_area
    ctx.obj['FRICTION_FPATH'] = friction_fpath
    ctx.obj['FRICTION_DSET'] = friction_dset
    ctx.obj['CAP_COST_SCALE'] = cap_cost_scale
    ctx.obj['OUT_DIR'] = out_dir
    ctx.obj['MAX_WORKERS'] = max_workers
    ctx.obj['SITES_PER_WORKER'] = sites_per_worker
    ctx.obj['LOG_DIR'] = log_dir
    ctx.obj['RECALC_LCOE'] = recalc_lcoe
    ctx.obj['PRE_EXTRACT_INCLUSIONS'] = pre_extract_inclusions
    ctx.obj['VERBOSE'] = verbose

    if ctx.invoked_subcommand is None:
        t0 = time.time()
        init_mult(name, log_dir, modules=[__name__, 'reV', 'rex'],
                  verbose=verbose)

        dsets = []
        paths = excl_fpath
        if isinstance(excl_fpath, str):
            paths = [excl_fpath]

        with MultiFileResource(paths) as res:
            dsets = res.dsets

        if tm_dset in dsets:
            logger.info('Found techmap "{}".'.format(tm_dset))
        elif tm_dset not in dsets and not isinstance(excl_fpath, str):
            msg = ('Could not find techmap dataset "{}" and cannot run '
                   'techmap with arbitrary multiple exclusion filepaths '
                   'to write to: {}'.format(tm_dset, excl_fpath))
            logger.error(msg)
            raise RuntimeError(msg)
        else:
            logger.info('Could not find techmap "{}". Running techmap module.'
                        .format(tm_dset))
            try:
                TechMapping.run(excl_fpath, res_fpath, dset=tm_dset)
            except Exception as e:
                msg = ('TechMapping process failed. Received the '
                       'following error:\n{}'.format(e))
                logger.exception(msg)
                raise RuntimeError(msg) from e

        if isinstance(excl_dict, str):
            excl_dict = dict_str_load(excl_dict)

        if isinstance(data_layers, str):
            data_layers = dict_str_load(data_layers)

        try:
            if gen_fpath is None:
                out = Aggregation.run(
                    excl_fpath, res_fpath, tm_dset,
                    excl_dict=excl_dict,
                    resolution=resolution,
                    excl_area=excl_area,
                    area_filter_kernel=area_filter_kernel,
                    min_area=min_area,
                    pre_extract_inclusions=pre_extract_inclusions,
                    max_workers=max_workers,
                    sites_per_worker=sites_per_worker)
                summary = out['meta']
            else:
                summary = SupplyCurveAggregation.summary(
                    excl_fpath, gen_fpath, tm_dset,
                    econ_fpath=econ_fpath,
                    excl_dict=excl_dict,
                    res_class_dset=res_class_dset,
                    res_class_bins=res_class_bins,
                    cf_dset=cf_dset,
                    lcoe_dset=lcoe_dset,
                    h5_dsets=h5_dsets,
                    data_layers=data_layers,
                    resolution=resolution,
                    excl_area=excl_area,
                    power_density=power_density,
                    area_filter_kernel=area_filter_kernel,
                    min_area=min_area,
                    friction_fpath=friction_fpath,
                    friction_dset=friction_dset,
                    cap_cost_scale=cap_cost_scale,
                    recalc_lcoe=recalc_lcoe,
                    pre_extract_inclusions=pre_extract_inclusions,
                    max_workers=max_workers,
                    sites_per_worker=sites_per_worker)

        except Exception as e:
            msg = ('Supply curve Aggregation failed. Received the '
                   'following error:\n{}'.format(e))
            logger.exception(msg)
            raise RuntimeError(msg) from e

        fn_out = '{}.csv'.format(name)
        fpath_out = os.path.join(out_dir, fn_out)
        summary.to_csv(fpath_out)

        runtime = (time.time() - t0) / 60
        logger.info('Supply curve aggregation complete. '
                    'Time elapsed: {:.2f} min. Target output dir: {}'
                    .format(runtime, out_dir))

        finput = [excl_fpath, gen_fpath]
        if res_fpath is not None:
            finput.append(res_fpath)

        # add job to reV status file.
        status = {'dirout': out_dir,
                  'fout': fn_out,
                  'job_status': 'successful',
                  'runtime': runtime,
                  'finput': finput,
                  'gen_fpath': gen_fpath,
                  'econ_fpath': econ_fpath,
                  'excl_fpath': excl_fpath,
                  'excl_dict': excl_dict,
                  'friction_fpath': friction_fpath,
                  'friction_dset': friction_dset,
                  'cf_dset': cf_dset,
                  'lcoe_dset': lcoe_dset,
                  'tm_dset': tm_dset,
                  'res_class_dset': res_class_dset,
                  'res_class_bins': res_class_bins,
                  'cap_cost_scale': cap_cost_scale,
                  'power_density': power_density,
                  'resolution': resolution,
                  'recalc_lcoe': recalc_lcoe,
                  'area_filter_kernel': area_filter_kernel,
                  'min_area': min_area}

        Status.make_job_file(out_dir, ModuleName.SUPPLY_CURVE_AGGREGATION,
                             name, status)


def get_node_cmd(name, excl_fpath, gen_fpath, econ_fpath, res_fpath, tm_dset,
                 excl_dict, res_class_dset, res_class_bins,
                 cf_dset, lcoe_dset, h5_dsets, data_layers, resolution,
                 excl_area, power_density, area_filter_kernel, min_area,
                 friction_fpath, friction_dset, cap_cost_scale,
                 out_dir, max_workers, sites_per_worker, log_dir,
                 recalc_lcoe, pre_extract_inclusions, verbose):
    """Get a CLI call command for the SC aggregation cli."""

    args = ['-exf {}'.format(SLURM.s(excl_fpath)),
            '-gf {}'.format(SLURM.s(gen_fpath)),
            '-ef {}'.format(SLURM.s(econ_fpath)),
            '-rf {}'.format(SLURM.s(res_fpath)),
            '-tm {}'.format(SLURM.s(tm_dset)),
            '-exd {}'.format(SLURM.s(excl_dict)),
            '-cd {}'.format(SLURM.s(res_class_dset)),
            '-cb {}'.format(SLURM.s(res_class_bins)),
            '-cf {}'.format(SLURM.s(cf_dset)),
            '-lc {}'.format(SLURM.s(lcoe_dset)),
            '-hd {}'.format(SLURM.s(h5_dsets)),
            '-d {}'.format(SLURM.s(data_layers)),
            '-r {}'.format(SLURM.s(resolution)),
            '-ea {}'.format(SLURM.s(excl_area)),
            '-pd {}'.format(SLURM.s(power_density)),
            '-afk {}'.format(SLURM.s(area_filter_kernel)),
            '-ma {}'.format(SLURM.s(min_area)),
            '-ff {}'.format(SLURM.s(friction_fpath)),
            '-fd {}'.format(SLURM.s(friction_dset)),
            '-cs {}'.format(SLURM.s(cap_cost_scale)),
            '-o {}'.format(SLURM.s(out_dir)),
            '-mw {}'.format(SLURM.s(max_workers)),
            '-spw {}'.format(SLURM.s(sites_per_worker)),
            '-ld {}'.format(SLURM.s(log_dir)),
            '-recalc {}'.format(SLURM.s(recalc_lcoe)),
            ]

    if pre_extract_inclusions:
        args.append('-pre')

    if verbose:
        args.append('-v')

    cmd = ('python -m reV.supply_curve.cli_sc_aggregation -n {} direct {}'
           .format(SLURM.s(name), ' '.join(args)))
    logger.debug('Creating the following command line call:\n\t{}'.format(cmd))

    return cmd


@direct.command()
@click.option('--alloc', '-a', required=True, type=STR,
              help='SLURM allocation account name.')
@click.option('--walltime', '-wt', default=1.0, type=float,
              show_default=True,
              help='SLURM walltime request in hours. Default is 1.0')
@click.option('--feature', '-l', default=None, type=STR,
              show_default=True,
              help=('Additional flags for SLURM job. Format is "--qos=high" '
                    'or "--depend=[state:job_id]". Default is None.'))
@click.option('--memory', '-mem', default=None, type=INT,
              show_default=True,
              help='SLURM node memory request in GB. Default is None')
@click.option('--module', '-mod', default=None, type=STR,
              show_default=True,
              help='Module to load')
@click.option('--conda_env', '-env', default=None, type=STR,
              show_default=True,
              help='Conda env to activate')
@click.option('--stdout_path', '-sout', default=None, type=STR,
              show_default=True,
              help='Subprocess standard output path. Default is in out_dir.')
@click.option('--sh_script', '-sh', default=None, type=STR,
              show_default=True,
              help='Extra shell script commands to run before the reV call.')
@click.pass_context
def slurm(ctx, alloc, walltime, feature, memory, module, conda_env,
          stdout_path, sh_script):
    """slurm (Eagle) submission tool for reV supply curve aggregation."""
    name = ctx.obj['NAME']
    excl_fpath = ctx.obj['EXCL_FPATH']
    gen_fpath = ctx.obj['GEN_FPATH']
    econ_fpath = ctx.obj['ECON_FPATH']
    res_fpath = ctx.obj['RES_FPATH']
    tm_dset = ctx.obj['TM_DSET']
    excl_dict = ctx.obj['EXCL_DICT']
    res_class_dset = ctx.obj['RES_CLASS_DSET']
    res_class_bins = ctx.obj['RES_CLASS_BINS']
    cf_dset = ctx.obj['CF_DSET']
    lcoe_dset = ctx.obj['LCOE_DSET']
    h5_dsets = ctx.obj['H5_DSETS']
    data_layers = ctx.obj['DATA_LAYERS']
    resolution = ctx.obj['RESOLUTION']
    excl_area = ctx.obj['EXCL_AREA']
    power_density = ctx.obj['POWER_DENSITY']
    area_filter_kernel = ctx.obj['AREA_FILTER_KERNEL']
    min_area = ctx.obj['MIN_AREA']
    friction_fpath = ctx.obj['FRICTION_FPATH']
    friction_dset = ctx.obj['FRICTION_DSET']
    cap_cost_scale = ctx.obj['CAP_COST_SCALE']
    out_dir = ctx.obj['OUT_DIR']
    max_workers = ctx.obj['MAX_WORKERS']
    sites_per_worker = ctx.obj['SITES_PER_WORKER']
    log_dir = ctx.obj['LOG_DIR']
    recalc_lcoe = ctx.obj['RECALC_LCOE']
    pre_extract_inclusions = ctx.obj['PRE_EXTRACT_INCLUSIONS']
    verbose = ctx.obj['VERBOSE']

    if stdout_path is None:
        stdout_path = os.path.join(log_dir, 'stdout/')

    cmd = get_node_cmd(name, excl_fpath, gen_fpath, econ_fpath, res_fpath,
                       tm_dset, excl_dict,
                       res_class_dset, res_class_bins,
                       cf_dset, lcoe_dset, h5_dsets, data_layers,
                       resolution, excl_area,
                       power_density, area_filter_kernel, min_area,
                       friction_fpath, friction_dset, cap_cost_scale,
                       out_dir, max_workers, sites_per_worker,
                       log_dir, recalc_lcoe,
                       pre_extract_inclusions, verbose)

    if sh_script:
        cmd = sh_script + '\n' + cmd

    slurm_manager = ctx.obj.get('SLURM_MANAGER', None)
    if slurm_manager is None:
        slurm_manager = SLURM()
        ctx.obj['SLURM_MANAGER'] = slurm_manager

    status = Status.retrieve_job_status(
        out_dir, module=ModuleName.SUPPLY_CURVE_AGGREGATION, job_name=name,
        hardware='slurm', subprocess_manager=slurm_manager)

    msg = 'SC Aggregation CLI failed to submit jobs!'
    if status == 'successful':
        msg = ('Job "{}" is successful in status json found in "{}", '
               'not re-running.'
               .format(name, out_dir))
    elif 'fail' not in str(status).lower() and status is not None:
        msg = ('Job "{}" was found with status "{}", not resubmitting'
               .format(name, status))
    else:
        logger.info('Running reV SC aggregation on SLURM with '
                    'node name "{}"'.format(name))
        out = slurm_manager.sbatch(cmd, alloc=alloc, memory=memory,
                                   walltime=walltime, feature=feature,
                                   name=name, stdout_path=stdout_path,
                                   conda_env=conda_env, module=module)[0]
        if out:
            msg = ('Kicked off reV SC aggregation job "{}" '
                   '(SLURM jobid #{}).'
                   .format(name, out))

        Status.add_job(
            out_dir, module=ModuleName.SUPPLY_CURVE_AGGREGATION,
            job_name=name, replace=True,
            job_attrs={'job_id': out, 'hardware': 'slurm',
                       'fout': '{}.csv'.format(name), 'dirout': out_dir})

    click.echo(msg)
    logger.info(msg)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running reV SC Aggregation CLI.')
        raise



# -*- coding: utf-8 -*-
"""
place turbines for bespoke wind plants
"""
import numpy as np
from shapely.geometry import Polygon, MultiPolygon
import shapely.affinity

from reV.bespoke.pack_turbs import PackTurbines
from reV.bespoke.gradient_free import GeneticAlgorithm
from reV.utilities.exceptions import WhileLoopPackingError


class PlaceTurbines():
    """Framework for optimizing turbine locations for site specific
    exclusions, wind resources, and objective
    """

    def __init__(self, wind_plant, objective_function,
                 capital_cost_function,
                 fixed_operating_cost_function,
                 variable_operating_cost_function,
                 include_mask, pixel_side_length, min_spacing,
                 wake_loss_multiplier=1):
        """
        Parameters
        ----------
        wind_plant : WindPowerPD
            wind plant object to analyze wind plant performance. This object
            should have everything in the plant defined, such that only the
            turbine coordinates and plant capacity need to be defined during
            the optimization.
        objective_function : str
            The objective function of the optimization as a string, should
            return the objective to be minimized during layout optimization.
            Variables available are:
                - n_turbines: the number of turbines
                - system_capacity: wind plant capacity
                - aep: annual energy production
                - fixed_charge_rate: user input fixed_charge_rate if included
                  as part of the sam system config.
                - capital_cost: plant capital cost as evaluated
                  by `capital_cost_function`
                - fixed_operating_cost: plant fixed annual operating cost as
                  evaluated by `fixed_operating_cost_function`
                - variable_operating_cost: plant variable annual operating cost
                  as evaluated by `variable_operating_cost_function`
                - self.wind_plant: the SAM wind plant object, through which
                  all SAM variables can be accessed
                - cost: the annual cost of the wind plant (from cost_function)
        capital_cost_function : str
            The plant capital cost function as a string, must return the total
            capital cost in $. Has access to the same variables as the
            objective_function.
        fixed_operating_cost_function : str
            The plant annual fixed operating cost function as a string, must
            return the fixed operating cost in $/year. Has access to the same
            variables as the objective_function.
        variable_operating_cost_function : str
            The plant annual variable operating cost function as a string, must
            return the variable operating cost in $/kWh. Has access to the same
            variables as the objective_function.
        exclusions : ExclusionMaskFromDict
            The exclusions that define where turbines can be placed. Contains
            exclusions.latitude, exclusions.longitude, and exclusions.mask
        min_spacing : float
            The minimum spacing between turbines (in meters).
        wake_loss_multiplier : float, optional
            A multiplier used to scale the annual energy lost due to
            wake losses. **IMPORTANT**: This multiplier will ONLY be
            applied during the optimization process and will NOT be
            come through in output values such as aep, any of the cost
            functions, or even the output objective.
        """

        # inputs
        self.wind_plant = wind_plant

        self.capital_cost_function = capital_cost_function
        self.fixed_operating_cost_function = fixed_operating_cost_function
        self.variable_operating_cost_function = \
            variable_operating_cost_function

        self.objective_function = objective_function
        self.include_mask = include_mask
        self.pixel_side_length = pixel_side_length
        self.min_spacing = min_spacing
        self.wake_loss_multiplier = wake_loss_multiplier

        # internal variables
        self.nrows, self.ncols = np.shape(include_mask)
        self.x_locations = np.array([])
        self.y_locations = np.array([])
        self.turbine_capacity = \
            np.max(self.wind_plant.
                   sam_sys_inputs["wind_turbine_powercurve_powerout"])
        self.full_polygons = None
        self.packing_polygons = None
        self.optimized_design_variables = None
        self.safe_polygons = None

        self.ILLEGAL = ('import ', 'os.', 'sys.', '.__', '__.', 'eval', 'exec')
        self._preflight(self.objective_function)
        self._preflight(self.capital_cost_function)
        self._preflight(self.fixed_operating_cost_function)
        self._preflight(self.variable_operating_cost_function)

    def _preflight(self, eqn):
        """Run preflight checks on the equation string."""
        for substr in self.ILLEGAL:
            if substr in str(eqn):
                msg = ('Will not evaluate string which contains "{}": {}'
                       .format(substr, eqn))
                raise ValueError(msg)

    def define_exclusions(self):
        """From the exclusions data, create a shapely MultiPolygon as
        self.safe_polygons that defines where turbines can be placed.
        """

        nx, ny = np.shape(self.include_mask)
        self.safe_polygons = MultiPolygon()
        side_x = np.linspace(0.0, float(nx), nx + 1)
        side_y = np.linspace(0.0, float(ny), ny + 1)
        floored = np.floor(self.include_mask)
        for i in range(nx):
            for j in range(ny):
                if floored[j, i] == 1:
                    added_poly = Polygon(((side_x[i], side_y[j]),
                                          (side_x[i + 1], side_y[j]),
                                          (side_x[i + 1], side_y[j + 1]),
                                          (side_x[i], side_y[j + 1])))
                    self.safe_polygons = self.safe_polygons.union(added_poly)

        self.safe_polygons = \
            shapely.affinity.scale(self.safe_polygons,
                                   xfact=self.pixel_side_length,
                                   yfact=-self.pixel_side_length,
                                   origin=(0, 0))

        if self.safe_polygons.area == 0.0:
            self.full_polygons = MultiPolygon([])
            self.packing_polygons = MultiPolygon([])
        else:
            minx, miny, maxx, maxy = self.safe_polygons.bounds
            self.safe_polygons = shapely.affinity.translate(self.safe_polygons,
                                                            xoff=-minx,
                                                            yoff=-miny)
            self.full_polygons = self.safe_polygons.buffer(0)

            # add extra setback to cell boundary
            minx, miny, maxx, maxy = self.full_polygons.bounds
            minx += self.min_spacing / 2.0
            miny += self.min_spacing / 2.0
            maxx -= self.min_spacing / 2.0
            maxy -= self.min_spacing / 2.0

            boundary_poly = \
                Polygon(((minx, miny), (minx, maxy), (maxx, maxy),
                         (maxx, miny)))
            packing_polygons = boundary_poly.intersection(self.full_polygons)
            if isinstance(packing_polygons, MultiPolygon):
                self.packing_polygons = packing_polygons
            elif isinstance(packing_polygons, Polygon):
                self.packing_polygons = MultiPolygon([packing_polygons])
            else:
                self.packing_polygons = MultiPolygon([])

    def initialize_packing(self):
        """run the turbine packing algorithm (maximizing plant capacity) to
        define potential turbine locations that will be used as design
        variables in the gentic algorithm.
        """

        packing = PackTurbines(self.min_spacing, self.packing_polygons)
        nturbs = 1E6
        mult = 1.0
        iters = 0
        while nturbs > 300:
            iters += 1
            if iters > 10000:
                msg = ('Too many attempts within initialize packing')
                raise WhileLoopPackingError(msg)
            packing.clear()
            packing.min_spacing = self.min_spacing * mult
            packing.pack_turbines_poly()
            nturbs = len(packing.turbine_x)
            mult *= 1.1
        self.x_locations = packing.turbine_x
        self.y_locations = packing.turbine_y

    # pylint: disable=W0641,W0123
    def optimization_objective(self, x):
        """The optimization objective used in the bespoke optimization
        """
        x = [bool(y) for y in x]
        n_turbines = np.sum(x)
        self.wind_plant["wind_farm_xCoordinates"] = self.x_locations[x]
        self.wind_plant["wind_farm_yCoordinates"] = self.y_locations[x]

        system_capacity = n_turbines * self.turbine_capacity
        self.wind_plant["system_capacity"] = system_capacity

        self.wind_plant.assign_inputs()
        self.wind_plant.execute()
        aep = self._aep_after_scaled_wake_losses()

        fixed_charge_rate = self.fixed_charge_rate
        capital_cost = eval(self.capital_cost_function,
                            globals(), locals())
        fixed_operating_cost = eval(self.fixed_operating_cost_function,
                                    globals(), locals())
        variable_operating_cost = eval(self.variable_operating_cost_function,
                                       globals(), locals())

        objective = eval(self.objective_function, globals(), locals())

        return objective

    def _aep_after_scaled_wake_losses(self):
        """AEP after scaling the energy lost due to wake."""
        wake_loss_pct = self.wind_plant['wake_losses']
        aep = self.wind_plant['annual_energy']
        agep = self.wind_plant['annual_gross_energy']

        energy_lost_due_to_wake = wake_loss_pct / 100 * agep
        aep_after_wake_losses = agep - energy_lost_due_to_wake
        other_losses_multiplier = 1 - aep / aep_after_wake_losses

        scaled_wake_losses = (self.wake_loss_multiplier
                              * energy_lost_due_to_wake)
        aep_after_scaled_wake_losses = max(0, agep - scaled_wake_losses)
        return aep_after_scaled_wake_losses * (1 - other_losses_multiplier)

    def optimize(self, **kwargs):
        """Optimize wind farm layout.

        Use a genetic algorithm to optimize wind plant layout for the
        user-defined objective function.

        Parameters
        ----------
        **kwargs
            Keyword arguments to pass to GA initialization.

        See Also
        --------
        :class:`~reV.bespoke.gradient_free.GeneticAlgorithm` : GA Algorithm.
        """
        nlocs = len(self.x_locations)
        bits = np.ones(nlocs, dtype=int)
        bounds = np.zeros((nlocs, 2), dtype=int)
        bounds[:, 1] = 2
        variable_type = np.array([])
        for _ in range(nlocs):
            variable_type = np.append(variable_type, "int")

        ga_kwargs = {
            'max_generation': 10000,
            'population_size': 25,
            'crossover_rate': 0.2,
            'mutation_rate': 0.01,
            'tol': 1E-6,
            'convergence_iters': 10000,
            'max_time': 3600
        }

        ga_kwargs.update(kwargs)

        ga = GeneticAlgorithm(bits, bounds, variable_type,
                              self.optimization_objective,
                              **ga_kwargs)

        ga.optimize_ga()

        optimized_design_variables = ga.optimized_design_variables
        self.optimized_design_variables = \
            [bool(y) for y in optimized_design_variables]

    def place_turbines(self, **kwargs):
        """Define bespoke wind plant turbine layouts.

        Run all functions to define bespoke wind plant turbine layouts.

        Parameters
        ----------
        **kwargs
            Keyword arguments to pass to GA initialization.

        See Also
        --------
        :class:`~reV.bespoke.gradient_free.GeneticAlgorithm` : GA Algorithm.
        """
        self.define_exclusions()
        self.initialize_packing()
        self.optimize(**kwargs)

    @property
    def turbine_x(self):
        """This is the final optimized turbine x locations (m)"""
        if self.optimized_design_variables is not None:
            return self.x_locations[self.optimized_design_variables]
        else:
            return None

    @property
    def turbine_y(self):
        """This is the final optimized turbine y locations (m)"""
        if self.optimized_design_variables is not None:
            return self.y_locations[self.optimized_design_variables]
        else:
            return None

    @property
    def nturbs(self):
        """This is the final optimized number of turbines"""
        if self.optimized_design_variables is not None:
            return np.sum(self.optimized_design_variables)
        else:
            return None

    @property
    def capacity(self):
        """This is the final optimized plant nameplate capacity (kW)"""
        if self.optimized_design_variables is not None:
            return self.turbine_capacity * self.nturbs
        else:
            return None

    @property
    def area(self):
        """This is the area available for wind turbine placement (km2)"""
        if self.full_polygons is not None:
            return self.full_polygons.area
        else:
            return None

    @property
    def fixed_charge_rate(self):
        """Fixed charge rate if input to the SAM WindPowerPD object, None if
        not found in inputs."""
        return self.wind_plant.sam_sys_inputs.get('fixed_charge_rate', None)

    @property
    def capacity_density(self):
        """This is the optimized capacity density of the wind plant
        defined with the area available after removing the exclusions
        (MW/km2)"""
        if self.full_polygons is None or self.capacity is None:
            return None
        else:
            if self.area != 0.0:
                return self.capacity / self.area * 1E3
            else:
                return 0.0

    @property
    def aep(self):
        """This is the annual energy production of the optimized plant (kWh)"""
        if self.optimized_design_variables is not None:
            self.wind_plant["wind_farm_xCoordinates"] = self.turbine_x
            self.wind_plant["wind_farm_yCoordinates"] = self.turbine_y
            self.wind_plant["system_capacity"] = self.capacity
            self.wind_plant.assign_inputs()
            self.wind_plant.execute()
            return self.wind_plant.annual_energy()
        else:
            return None

    # pylint: disable=W0641,W0123
    @property
    def capital_cost(self):
        """This is the capital cost of the optimized plant ($)"""
        if self.optimized_design_variables is not None:
            fixed_charge_rate = self.fixed_charge_rate
            n_turbines = self.nturbs
            system_capacity = self.capacity
            aep = self.aep
            return eval(self.capital_cost_function, globals(), locals())
        else:
            return None

    # pylint: disable=W0641,W0123
    @property
    def fixed_operating_cost(self):
        """This is the annual fixed operating cost of the
        optimized plant ($/year)"""
        if self.optimized_design_variables is not None:
            fixed_charge_rate = self.fixed_charge_rate
            n_turbines = self.nturbs
            system_capacity = self.capacity
            aep = self.aep
            return eval(self.fixed_operating_cost_function,
                        globals(), locals())
        else:
            return None

    # pylint: disable=W0641,W0123
    @property
    def variable_operating_cost(self):
        """This is the annual variable operating cost of the
        optimized plant ($/kWh)"""
        if self.optimized_design_variables is not None:
            fixed_charge_rate = self.fixed_charge_rate
            n_turbines = self.nturbs
            system_capacity = self.capacity
            aep = self.aep
            return eval(self.variable_operating_cost_function,
                        globals(), locals())
        else:
            return None

    # pylint: disable=W0641,W0123
    @property
    def objective(self):
        """This is the optimized objective function value"""
        if self.optimized_design_variables is not None:
            fixed_charge_rate = self.fixed_charge_rate
            n_turbines = self.nturbs
            system_capacity = self.capacity
            aep = self.aep
            capital_cost = self.capital_cost
            fixed_operating_cost = self.fixed_operating_cost
            variable_operating_cost = self.variable_operating_cost
            return eval(self.objective_function, globals(), locals())
        else:
            return None



# -*- coding: utf-8 -*-
"""reV-to-SAM interface module.

Wraps the NREL-PySAM library with additional reV features.
"""
import copy
import json
import logging
import numpy as np
import os
import pandas as pd
from warnings import warn
import PySAM.GenericSystem as generic

from reV.utilities.exceptions import (SAMInputWarning, SAMInputError,
                                      SAMExecutionError, ResourceError)

from rex.multi_file_resource import (MultiFileResource, MultiFileNSRDB,
                                     MultiFileWTK)
from rex.renewable_resource import (WindResource, SolarResource, NSRDB,
                                    WaveResource)
from rex.utilities.utilities import check_res_file


logger = logging.getLogger(__name__)


class SamResourceRetriever:
    """Factory utility to get the SAM resource handler."""

    @staticmethod
    def _get_base_handler(res_file, module):
        """Get the base SAM resource handler, raise error if module not found.

        Parameters
        ----------
        res_file : str
            Single resource file (with full path) to retrieve.
        module : str
            SAM module name or reV technology to force interpretation
            of the resource file type.
            Example: module set to 'pvwatts' or 'tcsmolten' means that this
            expects a SolarResource file. If 'nsrdb' is in the res_file name,
            the NSRDB handler will be used.

        Returns
        -------
        res_handler : SolarResource | WindResource | NSRDB
            Solar or Wind resource handler based on input.
        """

        try:
            res_handler = RevPySam.RESOURCE_TYPES[module.lower()]

        except KeyError as e:
            msg = ('Cannot interpret what kind of resource handler the SAM '
                   'module or reV technology "{}" requires. Expecting one of '
                   'the following SAM modules or reV technologies: {}'
                   .format(module, list(RevPySam.RESOURCE_TYPES.keys())))
            logger.exception(msg)
            raise SAMExecutionError(msg) from e

        if res_handler == SolarResource and 'nsrdb' in res_file.lower():
            # Use NSRDB handler if definitely an NSRDB file
            res_handler = NSRDB

        return res_handler

    @staticmethod
    def _parse_gid_map_sites(gen_gids, gid_map=None):
        """Parse resource gids based on the generation gids used by
        project_points and a gid_map. If gid_map is None, the input gen_gids
        are just passed through as the res_gids.

        Parameters
        ----------
        gen_gids : list
            List of project_points "sites" that are the generation gids.
        gid_map : None | dict
            Mapping of unique integer generation gids (keys) to single integer
            resource gids (values). This enables the user to input unique
            generation gids in the project points that map to non-unique
            resource gids. This can be None or a pre-extracted dict.

        Returns
        -------
        res_gids : list
            List of resource gids corresponding to the generation gids used by
            project points. If gid_map is None, then this is the same as the
            input gen_gids.
        """
        if gid_map is None:
            res_gids = gen_gids
        else:
            res_gids = [gid_map[i] for i in gen_gids]
        return res_gids

    @classmethod
    def _make_res_kwargs(cls, res_handler, project_points, output_request,
                         gid_map):
        """
        Make Resource.preloadSam args and kwargs

        Parameters
        ----------
        res_handler : Resource handler
            Wind resource handler.
        project_points : reV.config.ProjectPoints
            reV Project Points instance used to retrieve resource data at a
            specific set of sites.
        output_request : list
            Outputs to retrieve from SAM.
        gid_map : None | dict
            Mapping of unique integer generation gids (keys) to single integer
            resource gids (values). This enables the user to input unique
            generation gids in the project points that map to non-unique
            resource gids. This can be None or a pre-extracted dict.

        Returns
        -------
        kwargs : dict
            Extra input args to preload sam resource.
        args : tuple
            Args for res_handler.preload_SAM class method
        """
        sites = cls._parse_gid_map_sites(project_points.sites, gid_map=gid_map)
        args = (sites,)

        kwargs = {}
        if res_handler in (SolarResource, NSRDB):
            # check for clearsky irradiation analysis for NSRDB
            kwargs['clearsky'] = project_points.sam_config_obj.clearsky
            kwargs['bifacial'] = project_points.sam_config_obj.bifacial
            kwargs['tech'] = project_points.tech

            downscale = project_points.sam_config_obj.downscale
            # check for downscaling request
            if downscale is not None:
                # make sure that downscaling is only requested for NSRDB
                # resource
                if res_handler != NSRDB:
                    msg = ('Downscaling was requested for a non-NSRDB '
                           'resource file. reV does not have this capability '
                           'at the current time. Please contact a developer '
                           'for more information on this feature.')
                    logger.warning(msg)
                    warn(msg, SAMInputWarning)
                else:
                    # pass through the downscaling request
                    kwargs['downscale'] = downscale

        elif res_handler == WindResource:
            args += (project_points.h, )
            kwargs['icing'] = project_points.sam_config_obj.icing
            if project_points.curtailment is not None:
                if project_points.curtailment.precipitation:
                    # make precip rate available for curtailment analysis
                    kwargs['precip_rate'] = True

        # Check for resource means
        if any(req.endswith('_mean') for req in output_request):
            kwargs['means'] = True

        return kwargs, args

    @staticmethod
    def _multi_file_mods(res_handler, kwargs, res_file):
        """
        Check if res_file is a multi-file resource dir and update handler

        Parameters
        ----------
        res_handler : Resource
            Resource handler.
        kwargs : dict
            Key word arguments for resource init.
        res_file : str
            Single resource file (with full path) or multi h5 dir.

        Returns
        -------
        res_handler : Resource | MultiFileResource
            Resource handler, replaced by the multi file resource handler if
            necessary.
        kwargs : dict
            Key word arguments for resource init with h5_dir, prefix,
            and suffix.
        res_file : str
            Single resource file (with full path) or multi h5 dir.
        """
        if res_handler == WindResource:
            res_handler = MultiFileWTK
        elif res_handler == NSRDB:
            res_handler = MultiFileNSRDB
        else:
            res_handler = MultiFileResource

        return res_handler, kwargs, res_file

    @classmethod
    def get(cls, res_file, project_points, module,
            output_request=('cf_mean', ), gid_map=None):
        """Get the SAM resource iterator object (single year, single file).

        Parameters
        ----------
        res_file : str
            Single resource file (with full path) to retrieve.
        project_points : reV.config.ProjectPoints
            reV Project Points instance used to retrieve resource data at a
            specific set of sites.
        module : str
            SAM module name or reV technology to force interpretation
            of the resource file type.
            Example: module set to 'pvwatts' or 'tcsmolten' means that this
            expects a SolarResource file. If 'nsrdb' is in the res_file name,
            the NSRDB handler will be used.
        output_request : list | tuple, optional
            Outputs to retrieve from SAM, by default ('cf_mean', )
        gid_map : None | dict
            Mapping of unique integer generation gids (keys) to single integer
            resource gids (values). This enables the user to input unique
            generation gids in the project points that map to non-unique
            resource gids. This can be None or a pre-extracted dict.

        Returns
        -------
        res : reV.resource.SAMResource
            Resource iterator object to pass to SAM.
        """

        res_handler = cls._get_base_handler(res_file, module)
        kwargs, args = cls._make_res_kwargs(res_handler, project_points,
                                            output_request, gid_map)

        multi_h5_res, hsds = check_res_file(res_file)
        if multi_h5_res:
            res_handler, kwargs, res_file = cls._multi_file_mods(res_handler,
                                                                 kwargs,
                                                                 res_file)
        else:
            kwargs['hsds'] = hsds

        kwargs['time_index_step'] = \
            project_points.sam_config_obj.time_index_step

        res = res_handler.preload_SAM(res_file, *args, **kwargs)

        return res


class Sam:
    """reV wrapper on the PySAM framework."""

    # PySAM object wrapped by this class
    PYSAM = generic

    # callable attributes to be ignored in the get/set logic
    IGNORE_ATTRS = ['assign', 'execute', 'export']

    def __init__(self):
        self._pysam = self.PYSAM.new()
        self._attr_dict = None
        self._inputs = []
        self.sam_sys_inputs = {}
        if 'constant' in self.input_list:
            self['constant'] = 0.0

    def __getitem__(self, key):
        """Get the value of a PySAM attribute (either input or output).

        Parameters
        ----------
        key : str
            Lowest level attribute name.

        Returns
        -------
        out : object
            PySAM data.
        """

        group = self._get_group(key)
        try:
            out = getattr(getattr(self.pysam, group), key)
        except Exception:
            out = None

        return out

    def __setitem__(self, key, value):
        """Set a PySAM input data attribute.

        Parameters
        ----------
        key : str
            Lowest level attribute name.
        value : object
            Data to set to the key.
        """

        if key not in self.input_list:
            msg = ('Could not set input key "{}". Attribute not '
                   'found in PySAM object: "{}"'
                   .format(key, self.pysam))
            logger.exception(msg)
            raise SAMInputError(msg)
        else:
            self.sam_sys_inputs[key] = value
            group = self._get_group(key, outputs=False)
            try:
                setattr(getattr(self.pysam, group), key, value)
            except Exception as e:
                msg = ('Could not set input key "{}" to '
                       'group "{}" in "{}".\n'
                       'Data is: {} ({})\n'
                       'Received the following error: "{}"'
                       .format(key, group, self.pysam, value, type(value), e))
                logger.exception(msg)
                raise SAMInputError(msg) from e

    @property
    def pysam(self):
        """Get the pysam object."""
        return self._pysam

    @classmethod
    def default(cls):
        """Get the executed default pysam object.

        Returns
        -------
        PySAM.GenericSystem
        """
        obj = cls.PYSAM.default('GenericSystemNone')
        obj.execute()

        return obj

    @property
    def attr_dict(self):
        """Get the heirarchical PySAM object attribute dictionary.

        Returns
        -------
        _attr_dict : dict
            Dictionary with:
               keys: variable groups
               values: lowest level attribute/variable names
        """
        if self._attr_dict is None:
            keys = self._get_pysam_attrs(self.pysam)
            self._attr_dict = {k: self._get_pysam_attrs(getattr(self.pysam, k))
                               for k in keys}

        return self._attr_dict

    @property
    def input_list(self):
        """Get the list of lowest level input attribute/variable names.

        Returns
        -------
        _inputs : list
            List of lowest level input attributes.
        """
        if not any(self._inputs):
            for k, v in self.attr_dict.items():
                if k.lower() != 'outputs':
                    self._inputs += v

        return self._inputs

    def _get_group(self, key, outputs=True):
        """Get the group that the input key belongs to.

        Parameters
        ----------
        key : str
            Lowest level PySAM attribute/variable name.
        outputs : bool
            Flag if this key might be in outputs group. False ignores the
            outputs group (looks for inputs only).

        Returns
        -------
        group : str | None
            PySAM attribute group that key belongs to. None if not found.
        """
        group = None

        temp = self.attr_dict
        if not outputs:
            temp = {k: v for (k, v) in temp.items()
                    if k.lower() != 'outputs'}

        for k, v in temp.items():
            if key in v:
                group = k
                break

        return group

    def _get_pysam_attrs(self, obj):
        """Get a list of attributes from obj with ignore logic.

        Parameters
        ----------
        obj : PySAM object
            PySAM object to get attribute list from.

        Returns
        -------
        attrs : list
            List of attrs belonging to obj with dunder attrs and IGNORE_ATTRS
            not included.
        """
        attrs = [a for a in dir(obj) if not a.startswith('__')
                 and a not in self.IGNORE_ATTRS]
        return attrs

    def execute(self):
        """Call the PySAM execute method. Raise SAMExecutionError if error."""
        try:
            self.pysam.execute()
        except Exception as e:
            msg = 'PySAM raised an error while executing: "{}"'.format(e)
            logger.exception(msg)
            raise SAMExecutionError(msg) from e

    @staticmethod
    def _filter_inputs(key, value):
        """Perform any necessary filtering of input keys and values for PySAM.

        Parameters
        ----------
        key : str
            SAM input key.
        value : str | int | float | list | np.ndarray
            Input value associated with key.

        Returns
        -------
        key : str
            Filtered SAM input key.
        value : str | int | float | list | np.ndarray
            Filtered Input value associated with key.
        """

        if '.' in key:
            key = key.replace('.', '_')

        if ':constant' in key and 'adjust:' in key:
            key = key.replace('adjust:', '')

        if isinstance(value, str) and '[' in value and ']' in value:
            value = json.loads(value)

        return key, value

    def assign_inputs(self, inputs, raise_warning=False):
        """Assign a flat dictionary of inputs to the PySAM object.

        Parameters
        ----------
        inputs : dict
            Flat (single-level) dictionary of PySAM inputs.
        raise_warning : bool
            Flag to raise a warning for inputs that are not set because they
            are not found in the PySAM object.
        """

        for k, v in inputs.items():
            k, v = self._filter_inputs(k, v)
            if k in self.input_list and v is not None:
                self[k] = v
            elif raise_warning:
                wmsg = ('Not setting input "{}" to: {}.'
                        .format(k, v))
                warn(wmsg, SAMInputWarning)
                logger.warning(wmsg)


class RevPySam(Sam):
    """Base class for reV-SAM simulations (generation and econ)."""

    DIR = os.path.dirname(os.path.realpath(__file__))
    MODULE = None

    # Mapping for reV technology and SAM module to h5 resource handler type
    # SolarResource is swapped for NSRDB if the res_file contains "nsrdb"
    RESOURCE_TYPES = {'pvwattsv5': SolarResource,
                      'pvwattsv7': SolarResource,
                      'pvsamv1': SolarResource,
                      'tcsmoltensalt': SolarResource,
                      'solarwaterheat': SolarResource,
                      'troughphysicalheat': SolarResource,
                      'lineardirectsteam': SolarResource,
                      'windpower': WindResource,
                      'mhkwave': WaveResource
                      }

    def __init__(self, meta, sam_sys_inputs, output_request,
                 site_sys_inputs=None):
        """Initialize a SAM object.

        Parameters
        ----------
        meta : pd.DataFrame
            1D table with resource meta data.
        sam_sys_inputs : dict
            Site-agnostic SAM system model inputs arguments.
        output_request : list
            Requested SAM outputs (e.g., 'cf_mean', 'annual_energy',
            'cf_profile', 'gen_profile', 'energy_yield', 'ppa_price',
            'lcoe_fcr').
        site_sys_inputs : dict
            Optional set of site-specific SAM system inputs to complement the
            site-agnostic inputs.
        """

        super().__init__()
        self._meta = meta
        self._site = None
        self.time_interval = 1
        self.outputs = {}
        self.sam_sys_inputs = sam_sys_inputs
        self.site_sys_inputs = site_sys_inputs
        self.output_request = output_request
        if self.output_request is None:
            self.output_request = []

        self._parse_site_sys_inputs(site_sys_inputs)

    @property
    def meta(self):
        """Get meta data property."""
        return self._meta

    @property
    def module(self):
        """Get module property."""
        return self.MODULE

    @property
    def site(self):
        """Get the site number for this SAM simulation."""
        return self._site

    @staticmethod
    def get_sam_res(*args, **kwargs):
        """Get the SAM resource iterator object (single year, single file)."""
        return SamResourceRetriever.get(*args, **kwargs)

    @staticmethod
    def drop_leap(resource):
        """Drop Feb 29th from resource df with time index.

        Parameters
        ----------
        resource : pd.DataFrame
            Resource dataframe with an index containing a pandas
            time index object with month and day attributes.

        Returns
        -------
        resource : pd.DataFrame
            Resource dataframe with all February 29th timesteps removed.
        """

        if hasattr(resource, 'index'):
            if (hasattr(resource.index, 'month')
                    and hasattr(resource.index, 'day')):
                leap_day = ((resource.index.month == 2)
                            & (resource.index.day == 29))
                resource = resource.drop(resource.index[leap_day])

        return resource

    @staticmethod
    def ensure_res_len(arr, time_index):
        """
        Ensure time_index has a constant time-step and only covers 365 days
        (no leap days). If not remove last day

        Parameters
        ----------
        arr : ndarray
            Array to truncate if time_index has a leap day
        time_index : pandas.DatatimeIndex
            Time index associated with arr, used to check time-series
            frequency and number of days

        Returns
        -------
        arr : ndarray
            Truncated array of data such that there are 365 days
        """
        msg = ('A valid time_index must be supplied to ensure the proper '
               'resource length! Instead {} was supplied'
               .format(type(time_index)))
        assert isinstance(time_index, pd.DatetimeIndex)

        msg = ('arr length {} does not match time_index length {}!'
               .format(len(arr), len(time_index)))
        assert len(arr) == len(time_index)

        if time_index.is_leap_year.all():
            mask = time_index.month == 2
            mask &= time_index.day == 29
            if not mask.any():
                mask = time_index.month == 2
                mask &= time_index.day == 28
                s = np.where(mask)[0][-1]

                freq = pd.infer_freq(time_index[:s])
                msg = 'frequencies do not match before and after 2/29'
                assert freq == pd.infer_freq(time_index[s + 1:]), msg
            else:
                freq = pd.infer_freq(time_index)
        else:
            freq = pd.infer_freq(time_index)

        if freq is None:
            msg = ('Resource time_index does not have a consistent time-step '
                   '(frequency)!')
            logger.error(msg)
            raise ResourceError(msg)

        doy = time_index.dayofyear
        n_doy = len(doy.unique())

        if n_doy > 365:
            # Drop last day of year
            doy_max = doy.max()
            mask = doy != doy_max
            arr = arr[mask]

        return arr

    @staticmethod
    def make_datetime(series):
        """Ensure that pd series is a datetime series with dt accessor"""
        if not hasattr(series, 'dt'):
            series = pd.to_datetime(pd.Series(series))

        return series

    @classmethod
    def get_time_interval(cls, time_index):
        """Get the time interval.

        Parameters
        ----------
        time_index : pd.series
            Datetime series. Must have a dt attribute to access datetime
            properties (added using make_datetime method).

        Returns
        -------
        time_interval : int:
            This value is the number of indices over which an hour is counted.
            So if the timestep is 0.5 hours, time_interval is 2.
        """

        time_index = cls.make_datetime(time_index)
        x = time_index.dt.hour.diff()
        time_interval = 0

        # iterate through the hourly time diffs and count indices between flips
        for t in x[1:]:
            if t == 1.0:
                time_interval += 1
                break
            elif t == 0.0:
                time_interval += 1

        return int(time_interval)

    def _parse_site_sys_inputs(self, site_sys_inputs):
        """Parse site-specific parameters and add to parameter dict.

        Parameters
        ----------
        site_sys_inputs : dict
            Optional set of site-specific SAM system inputs to complement the
            site-agnostic inputs.
        """

        if site_sys_inputs is not None:
            for k, v in site_sys_inputs.items():
                if isinstance(v, float) and np.isnan(v):
                    pass
                else:
                    self.sam_sys_inputs[k] = v

    @staticmethod
    def _is_arr_like(val):
        """Returns true if SAM data is array-like. False if scalar."""
        if isinstance(val, (int, float, str)):
            return False
        else:
            try:
                len(val)
            except TypeError:
                return False
            else:
                return True

    @classmethod
    def _is_hourly(cls, val):
        """Returns true if SAM data is hourly or sub-hourly. False otherise."""
        if not cls._is_arr_like(val):
            return False
        else:
            L = len(val)
            return L >= 8760

    def outputs_to_utc_arr(self):
        """Convert array-like SAM outputs to UTC np.ndarrays"""
        if self.outputs is not None:
            for key, output in self.outputs.items():
                if self._is_arr_like(output):
                    output = np.asarray(output)

                    if output.dtype == np.float64:
                        output = output.astype(np.float32)
                    elif output.dtype == np.int64:
                        output = output.astype(np.int32)

                    if self._is_hourly(output):
                        output = np.roll(output, int(-1 * self.meta['timezone']
                                                     * self.time_interval))

                    self.outputs[key] = output

    def collect_outputs(self, output_lookup):
        """Collect SAM output_request, convert timeseries outputs to UTC, and
        save outputs to self.outputs property.

        Parameters
        ----------
        output_lookup : dict
            Lookup dictionary mapping output keys to special output methods.
        """
        bad_requests = []
        for req in self.output_request:
            if req in output_lookup:
                self.outputs[req] = output_lookup[req]()
            elif req in self.sam_sys_inputs:
                self.outputs[req] = self.sam_sys_inputs[req]
            else:
                try:
                    self.outputs[req] = getattr(self.pysam.Outputs, req)
                except AttributeError:
                    bad_requests.append(req)

        if any(bad_requests):
            msg = ('Could not retrieve outputs "{}" from PySAM object "{}".'
                   .format(bad_requests, self.pysam))
            logger.error(msg)
            raise SAMExecutionError(msg)

        self.outputs_to_utc_arr()

    def assign_inputs(self):
        """Assign the self.sam_sys_inputs attribute to the PySAM object."""
        super().assign_inputs(copy.deepcopy(self.sam_sys_inputs))

    def execute(self):
        """Call the PySAM execute method. Raise SAMExecutionError if error.
        Include the site index if available.
        """
        try:
            self.pysam.execute()
        except Exception as e:
            msg = ('PySAM raised an error while executing: "{}"'
                   .format(self.module))
            if self.site is not None:
                msg += ' for site {}'.format(self.site)
            logger.exception(msg)
            raise SAMExecutionError(msg) from e



# -*- coding: utf-8 -*-
# pylint: disable=all
"""
reV-NRWAL module command line interface (CLI).

This module runs reV data through the NRWAL compute library. This code was
first developed to use a custom offshore wind LCOE equation library but has
since been refactored to analyze any equation library in NRWAL.

Everything in this module operates on the spatiotemporal resolution of the reV
generation output file. This is usually the wind or solar resource resolution
but could be the supply curve resolution after representative profiles is run.
"""
import pprint
import os
import click
import logging
import time

from reV.config.nrwal_config import RevNrwalConfig
from reV.pipeline.status import Status
from reV.nrwal.nrwal import RevNrwal
from reV.utilities.cli_dtypes import SAMFILES
from reV.utilities import ModuleName
from reV import __version__

from rex.utilities.cli_dtypes import STR, INT, STRLIST
from rex.utilities.loggers import init_mult
from rex.utilities.hpc import SLURM
from rex.utilities.utilities import get_class_properties

logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('--name', '-n', default=os.path.basename(os.getcwd()),
              type=STR, show_default=True, help='reV NRWAL job name.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, name, verbose):
    """reV-NRWAL Command Line Interface"""
    ctx.ensure_object(dict)
    ctx.obj['NAME'] = name
    ctx.obj['VERBOSE'] = verbose


@main.command()
def valid_config_keys():
    """
    Echo the valid nrwal config keys
    """
    click.echo(', '.join(get_class_properties(RevNrwalConfig)))


@main.command()
@click.option('--config_file', '-c', required=True,
              type=click.Path(exists=True),
              help='reV-NRWAL configuration json file.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def from_config(ctx, config_file, verbose):
    """Run reV-NRWAL analysis from a config file."""
    # Instantiate the config object
    config = RevNrwalConfig(config_file)

    # take name from config
    name = ctx.obj['NAME'] = config.name

    # Enforce verbosity if logging level is specified in the config
    if config.log_level == logging.DEBUG:
        verbose = True

    # initialize loggers
    init_mult(name, config.log_directory, modules=[__name__, 'reV', 'rex'],
              verbose=verbose)

    # Initial log statements
    logger.info('Running reV-NRWAL analysis from config '
                'file: "{}"'.format(config_file))
    logger.info('Target output directory: "{}"'.format(config.dirout))
    logger.info('Target logging directory: "{}"'.format(config.log_directory))
    logger.debug('The full configuration input is as follows:\n{}'
                 .format(pprint.pformat(config, indent=4)))

    gen_fpaths = config.gen_fpath
    if isinstance(gen_fpaths, str):
        gen_fpaths = [gen_fpaths]

    for i, gen_fpath in enumerate(gen_fpaths):
        job_name = '{}_{}'.format(name, str(i).zfill(2))
        ctx.obj['NAME'] = job_name

        if config.execution_control.option == 'local':
            status = Status.retrieve_job_status(config.dirout,
                                                module=ModuleName.NRWAL,
                                                job_name=job_name)
            if status != 'successful':
                Status.add_job(
                    config.dirout, module=ModuleName.NRWAL,
                    job_name=job_name, replace=True,
                    job_attrs={'hardware': 'local',
                               'fout': '{}_nrwal.h5'.format(job_name),
                               'dirout': config.dirout,
                               'finput': gen_fpath})
                ctx.invoke(direct,
                           gen_fpath=gen_fpath,
                           site_data=config.site_data,
                           sam_files=config.sam_files,
                           nrwal_configs=config.nrwal_configs,
                           output_request=config.output_request,
                           save_raw=config.save_raw,
                           meta_gid_col=config.meta_gid_col,
                           site_meta_cols=config.site_meta_cols,
                           log_dir=config.log_directory,
                           verbose=verbose)

        elif config.execution_control.option in ('eagle', 'slurm'):
            ctx.obj['GEN_FPATH'] = gen_fpath
            ctx.obj['SITE_DATA'] = config.site_data
            ctx.obj['SAM_FILES'] = config.sam_files
            ctx.obj['NRWAL_CONFIGS'] = config.nrwal_configs
            ctx.obj['OUTPUT_REQUEST'] = config.output_request
            ctx.obj['SAVE_RAW'] = config.save_raw
            ctx.obj['META_GID_COL'] = config.meta_gid_col
            ctx.obj['SITE_META_COLS'] = config.site_meta_cols
            ctx.obj['OUT_DIR'] = config.dirout
            ctx.obj['LOG_DIR'] = config.log_directory
            ctx.obj['VERBOSE'] = verbose

            ctx.invoke(slurm,
                       alloc=config.execution_control.allocation,
                       memory=config.execution_control.memory,
                       walltime=config.execution_control.walltime,
                       feature=config.execution_control.feature,
                       module=config.execution_control.module,
                       conda_env=config.execution_control.conda_env,
                       sh_script=config.execution_control.sh_script)


@main.group(invoke_without_command=True)
@click.option('--gen_fpath', '-gf', type=STR, required=True,
              help='reV wind generation/econ output file. Anything in the '
              'output_request is added and/or manipulated in this file.')
@click.option('--site_data', '-sd', type=STR, required=True,
              help='Site-specific input data for NRWAL calculation. String '
              'should be a filepath that points to a csv. Rows match sites, '
              'columns are input keys. Need a "gid" column that corresponds '
              'to the "meta_gid_col" in the gen_fpath meta data and a '
              '"config" column that corresponds to the nrwal_configs input. '
              'Only sites with a gid in this files "gid" column will be run '
              'through NRWAL.')
@click.option('--sam_files', '-sf', required=True, type=SAMFILES,
              help='SAM config files lookup mapping config keys to config '
              'filepaths. (required) (dict). Should have the same config '
              'keys as the nrwal_configs input.')
@click.option('--nrwal_configs', '-nc', required=True, type=SAMFILES,
              help='NRWAL config files lookup mapping config keys to config '
              'filepaths. (required) (dict). Should have the same config '
              'keys as the sam_files input.')
@click.option('--output_request', '-or', default=None, type=STRLIST,
              help='List of output dataset names you want written to the '
              'gen_fpath file. Any key from the NRWAL configs or any of the '
              'inputs (site_data or sam_files) is available to be exported as '
              'an output dataset. If you want to manipulate a dset like '
              'cf_mean from gen_fpath and include it in the output_request, '
              'you should set save_raw=True and then in the NRWAL equations '
              'use cf_mean_raw as the input and then define cf_mean as the '
              'manipulated data that will be included in the output_request.')
@click.option('--save_raw', '-sr', default=True, type=bool,
              help='Flag to save a copy of existing datasets in gen_fpath '
              'that are part of the output_request. For example, if you '
              'request cf_mean in output_request and manipulate the cf_mean '
              'dataset in the NRWAL equations, the original cf_mean will be '
              'archived under the "cf_mean_raw" dataset in gen_fpath.')
@click.option('--meta_gid_col', '-mg', default='gid', type=str,
              help='Column label in the source meta data from gen_fpath that '
              'contains the unique gid identifier. This will be joined to the '
              'site_data "gid" column.')
@click.option('--site_meta_cols', '-mc', default=None, type=STRLIST,
              help='Column labels from site_data to pass through to the '
              'output meta data. None (default) will use class variable '
              'DEFAULT_META_COLS, and any additional cols requested here will '
              'be added to DEFAULT_META_COLS.')
@click.option('--log_dir', '-ld', type=STR, default='./logs/',
              help='Directory to save reV-NRWAL logs.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def direct(ctx, gen_fpath, site_data, sam_files, nrwal_configs,
           output_request, save_raw, meta_gid_col,
           site_meta_cols, log_dir, verbose):
    """Main entry point to run reV-NRWAL analysis"""
    name = ctx.obj['NAME']
    ctx.obj['GEN_FPATH'] = gen_fpath
    ctx.obj['SITE_DATA'] = site_data
    ctx.obj['SAM_FILES'] = sam_files
    ctx.obj['NRWAL_CONFIGS'] = nrwal_configs
    ctx.obj['OUTPUT_REQUEST'] = output_request
    ctx.obj['SAVE_RAW'] = save_raw
    ctx.obj['META_GID_COL'] = meta_gid_col
    ctx.obj['SITE_META_COLS'] = site_meta_cols
    ctx.obj['LOG_DIR'] = log_dir
    ctx.obj['VERBOSE'] = verbose

    if ctx.invoked_subcommand is None:
        t0 = time.time()
        init_mult(name, log_dir, modules=[__name__, 'reV', 'rex'],
                  verbose=verbose, node=True)

        try:
            RevNrwal.run(gen_fpath, site_data, sam_files, nrwal_configs,
                         output_request,
                         save_raw=save_raw,
                         meta_gid_col=meta_gid_col,
                         site_meta_cols=site_meta_cols)
        except Exception as e:
            logger.exception('reV-NRWAL module failed, received the '
                             'following exception:\n{}'.format(e))
            raise e

        runtime = (time.time() - t0) / 60

        status = {'dirout': os.path.dirname(gen_fpath),
                  'fout': os.path.basename(gen_fpath),
                  'job_status': 'successful',
                  'runtime': runtime, 'finput': gen_fpath}
        Status.make_job_file(os.path.dirname(gen_fpath), ModuleName.NRWAL,
                             name, status)


def get_node_cmd(name, gen_fpath, site_data, sam_files, nrwal_configs,
                 output_request, save_raw, meta_gid_col, site_meta_cols,
                 log_dir, verbose):
    """Get a CLI call command for the reV-NRWAL cli."""

    args = ['-gf {}'.format(SLURM.s(gen_fpath)),
            '-sd {}'.format(SLURM.s(site_data)),
            '-sf {}'.format(SLURM.s(sam_files)),
            '-nc {}'.format(SLURM.s(nrwal_configs)),
            '-or {}'.format(SLURM.s(output_request)),
            '-sr {}'.format(SLURM.s(save_raw)),
            '-mg {}'.format(SLURM.s(meta_gid_col)),
            '-mc {}'.format(SLURM.s(site_meta_cols)),
            '-ld {}'.format(SLURM.s(log_dir)),
            ]

    if verbose:
        args.append('-v')

    cmd = ('python -m reV.nrwal.cli_nrwal -n {} direct {}'
           .format(SLURM.s(name), ' '.join(args)))
    logger.debug('Creating the following command line call:\n\t{}'.format(cmd))

    return cmd


@direct.command()
@click.option('--alloc', '-a', required=True, type=STR,
              help='SLURM allocation account name.')
@click.option('--feature', '-l', default=None, type=STR,
              help=('Additional flags for SLURM job. Format is "--qos=high" '
                    'or "--depend=[state:job_id]". Default is None.'))
@click.option('--memory', '-mem', default=None, type=INT, help='SLURM node '
              'memory request in GB. Default is None')
@click.option('--walltime', '-wt', default=1.0, type=float,
              help='SLURM walltime request in hours. Default is 1.0')
@click.option('--module', '-mod', default=None, type=STR,
              help='Module to load')
@click.option('--conda_env', '-env', default=None, type=STR,
              help='Conda env to activate')
@click.option('--stdout_path', '-sout', default=None, type=STR,
              help='Subprocess standard output path. Default is in out_dir.')
@click.option('--sh_script', '-sh', default=None, type=STR,
              show_default=True,
              help='Extra shell script commands to run before the reV call.')
@click.pass_context
def slurm(ctx, alloc, feature, memory, walltime, module, conda_env,
          stdout_path, sh_script):
    """slurm (Eagle) submission tool for reV supply curve aggregation."""

    name = ctx.obj['NAME']
    gen_fpath = ctx.obj['GEN_FPATH']
    site_data = ctx.obj['SITE_DATA']
    sam_files = ctx.obj['SAM_FILES']
    nrwal_configs = ctx.obj['NRWAL_CONFIGS']
    output_request = ctx.obj['OUTPUT_REQUEST']
    save_raw = ctx.obj['SAVE_RAW']
    meta_gid_col = ctx.obj['META_GID_COL']
    site_meta_cols = ctx.obj['SITE_META_COLS']
    out_dir = ctx.obj['OUT_DIR']
    log_dir = ctx.obj['LOG_DIR']
    verbose = ctx.obj['VERBOSE']

    if stdout_path is None:
        stdout_path = os.path.join(log_dir, 'stdout/')

    cmd = get_node_cmd(name, gen_fpath, site_data, sam_files, nrwal_configs,
                       output_request, save_raw, meta_gid_col, site_meta_cols,
                       log_dir, verbose)
    if sh_script:
        cmd = sh_script + '\n' + cmd

    slurm_manager = ctx.obj.get('SLURM_MANAGER', None)
    if slurm_manager is None:
        slurm_manager = SLURM()
        ctx.obj['SLURM_MANAGER'] = slurm_manager

    status = Status.retrieve_job_status(out_dir, module=ModuleName.NRWAL,
                                        job_name=name, hardware='slurm',
                                        subprocess_manager=slurm_manager)

    msg = 'NRWAL CLI failed to submit jobs!'
    if status == 'successful':
        msg = ('Job "{}" is successful in status json found in "{}", '
               'not re-running.'
               .format(name, out_dir))
    elif 'fail' not in str(status).lower() and status is not None:
        msg = ('Job "{}" was found with status "{}", not resubmitting'
               .format(name, status))
    else:
        logger.info('Running reV-NRWAL on SLURM with '
                    'node name "{}"'.format(name))
        out = slurm_manager.sbatch(cmd,
                                   alloc=alloc,
                                   memory=memory,
                                   walltime=walltime,
                                   feature=feature,
                                   name=name,
                                   stdout_path=stdout_path,
                                   conda_env=conda_env,
                                   module=module)[0]
        if out:
            msg = ('Kicked off reV-NRWAL job "{}" (SLURM jobid #{}).'
                   .format(name, out))

        Status.add_job(
            out_dir, module=ModuleName.NRWAL, job_name=name, replace=True,
            job_attrs={'job_id': out, 'hardware': 'slurm',
                       'fout': '{}.csv'.format(name), 'dirout': out_dir})

    click.echo(msg)
    logger.info(msg)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running reV-NRWAL CLI.')
        raise



# -*- coding: utf-8 -*-
"""
reV supply curve extent
"""
import logging
import numpy as np
import pandas as pd

from reV.handlers.exclusions import ExclusionLayers
from reV.utilities.exceptions import SupplyCurveError, SupplyCurveInputError

from rex.utilities.utilities import get_chunk_ranges

logger = logging.getLogger(__name__)


class SupplyCurveExtent:
    """Supply curve full extent framework. This class translates the 90m
    exclusion grid to the aggregated supply curve resolution."""

    def __init__(self, f_excl, resolution=64):
        """
        Parameters
        ----------
        f_excl : str | list | tuple | ExclusionLayers
            File path(s) to the exclusions grid, or pre-initialized
            ExclusionLayers. The exclusions dictate the SC analysis extent.
        resolution : int
            Number of exclusion points per SC point along an axis.
            This number**2 is the total number of exclusion points per
            SC point.
        """

        logger.debug('Initializing SupplyCurveExtent with res {} from: {}'
                     .format(resolution, f_excl))

        if not isinstance(resolution, int):
            raise SupplyCurveInputError('Supply Curve resolution needs to be '
                                        'an integer but received: {}'
                                        .format(type(resolution)))

        if isinstance(f_excl, (str, list, tuple)):
            self._excl_fpath = f_excl
            self._excls = ExclusionLayers(f_excl)
        elif isinstance(f_excl, ExclusionLayers):
            self._excl_fpath = f_excl.h5_file
            self._excls = f_excl
        else:
            raise SupplyCurveInputError('SupplyCurvePoints needs an '
                                        'exclusions file path, or '
                                        'ExclusionLayers handler but '
                                        'received: {}'
                                        .format(type(f_excl)))

        self._excl_shape = self.exclusions.shape
        # limit the resolution to the exclusion shape.
        self._res = int(np.min(list(self.excl_shape) + [resolution]))

        self._n_rows = None
        self._n_cols = None
        self._cols_of_excl = None
        self._rows_of_excl = None
        self._excl_row_slices = None
        self._excl_col_slices = None
        self._latitude = None
        self._longitude = None
        self._points = None

        self._sc_col_ind, self._sc_row_ind = np.meshgrid(
            np.arange(self.n_cols), np.arange(self.n_rows))
        self._sc_col_ind = self._sc_col_ind.flatten()
        self._sc_row_ind = self._sc_row_ind.flatten()

        logger.debug('Initialized SupplyCurveExtent with shape {} from '
                     'exclusions with shape {}'
                     .format(self.shape, self.excl_shape))

    def __len__(self):
        """Total number of supply curve points."""
        return self.n_rows * self.n_cols

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()
        if type is not None:
            raise

    def __getitem__(self, gid):
        """Get SC extent meta data corresponding to an SC point gid."""
        if gid >= len(self):
            raise KeyError('SC extent with {} points does not contain SC '
                           'point gid {}.'.format(len(self), gid))

        return self.points.loc[gid]

    def close(self):
        """Close all file handlers."""
        self._excls.close()

    @property
    def shape(self):
        """Get the Supply curve shape tuple (n_rows, n_cols).

        Returns
        -------
        shape : tuple
            2-entry tuple representing the full supply curve extent.
        """

        return (self.n_rows, self.n_cols)

    @property
    def exclusions(self):
        """Get the exclusions object.

        Returns
        -------
        _excls : ExclusionLayers
            ExclusionLayers h5 handler object.
        """
        return self._excls

    @property
    def resolution(self):
        """Get the 1D resolution.

        Returns
        -------
        _res : int
            Number of exclusion points per SC point along an axis.
            This number**2 is the total number of exclusion points per
            SC point.
        """
        return self._res

    @property
    def excl_shape(self):
        """Get the shape tuple of the exclusion file raster.

        Returns
        -------
        tuple
        """
        return self._excl_shape

    @property
    def excl_rows(self):
        """Get the unique row indices identifying the exclusion points.

        Returns
        -------
        excl_rows : np.ndarray
            Array of exclusion row indices.
        """
        return np.arange(self.excl_shape[0])

    @property
    def excl_cols(self):
        """Get the unique column indices identifying the exclusion points.

        Returns
        -------
        excl_cols : np.ndarray
            Array of exclusion column indices.
        """
        return np.arange(self.excl_shape[1])

    @property
    def rows_of_excl(self):
        """List representing the supply curve points rows and which
        exclusions rows belong to each supply curve row.

        Returns
        -------
        _rows_of_excl : list
            List representing the supply curve points rows. Each list entry
            contains the exclusion row indices that are included in the sc
            point.
        """
        if self._rows_of_excl is None:
            self._rows_of_excl = self._chunk_excl(self.excl_rows,
                                                  self.resolution)

        return self._rows_of_excl

    @property
    def cols_of_excl(self):
        """List representing the supply curve points columns and which
        exclusions columns belong to each supply curve column.

        Returns
        -------
        _cols_of_excl : list
            List representing the supply curve points columns. Each list entry
            contains the exclusion column indices that are included in the sc
            point.
        """
        if self._cols_of_excl is None:
            self._cols_of_excl = self._chunk_excl(self.excl_cols,
                                                  self.resolution)

        return self._cols_of_excl

    @property
    def excl_row_slices(self):
        """
        List representing the supply curve points rows and which
        exclusions rows belong to each supply curve row.

        Returns
        -------
        _excl_row_slices : list
            List representing the supply curve points rows. Each list entry
            contains the exclusion row slice that are included in the sc
            point.
        """
        if self._excl_row_slices is None:
            self._excl_row_slices = self._excl_slices(self.excl_rows,
                                                      self.resolution)

        return self._excl_row_slices

    @property
    def excl_col_slices(self):
        """
        List representing the supply curve points cols and which
        exclusions cols belong to each supply curve col.

        Returns
        -------
        _excl_col_slices : list
            List representing the supply curve points cols. Each list entry
            contains the exclusion col slice that are included in the sc
            point.
        """
        if self._excl_col_slices is None:
            self._excl_col_slices = self._excl_slices(self.excl_cols,
                                                      self.resolution)

        return self._excl_col_slices

    @property
    def n_rows(self):
        """Get the number of supply curve grid rows.

        Returns
        -------
        n_rows : int
            Number of row entries in the full supply curve grid.
        """
        if self._n_rows is None:
            self._n_rows = int(np.ceil(self.excl_shape[0] / self.resolution))

        return self._n_rows

    @property
    def n_cols(self):
        """Get the number of supply curve grid columns.

        Returns
        -------
        n_cols : int
            Number of column entries in the full supply curve grid.
        """
        if self._n_cols is None:
            self._n_cols = int(np.ceil(self.excl_shape[1] / self.resolution))

        return self._n_cols

    @property
    def latitude(self):
        """
        Get supply curve point latitudes

        Returns
        -------
        ndarray
        """
        if self._latitude is None:
            lats = []
            lons = []

            sc_cols, sc_rows = np.meshgrid(np.arange(self.n_cols),
                                           np.arange(self.n_rows))
            for r, c in zip(sc_rows.flatten(), sc_cols.flatten()):
                r = self.excl_row_slices[r]
                c = self.excl_col_slices[c]
                lats.append(self.exclusions['latitude', r, c].mean())
                lons.append(self.exclusions['longitude', r, c].mean())

            self._latitude = np.array(lats, dtype='float32')
            self._longitude = np.array(lons, dtype='float32')

        return self._latitude

    @property
    def longitude(self):
        """
        Get supply curve point longitudes

        Returns
        -------
        ndarray
        """
        if self._longitude is None:
            lats = []
            lons = []

            sc_cols, sc_rows = np.meshgrid(np.arange(self.n_cols),
                                           np.arange(self.n_rows))
            for r, c in zip(sc_rows.flatten(), sc_cols.flatten()):
                r = self.excl_row_slices[r]
                c = self.excl_col_slices[c]
                lats.append(self.exclusions['latitude', r, c].mean())
                lons.append(self.exclusions['longitude', r, c].mean())

            self._latitude = np.array(lats, dtype='float32')
            self._longitude = np.array(lons, dtype='float32')

        return self._longitude

    @property
    def lat_lon(self):
        """
        2D array of lat, lon coordinates for all sc points

        Returns
        -------
        ndarray
        """
        return np.dstack((self.latitude, self.longitude))[0]

    @property
    def row_indices(self):
        """Get a 1D array of row indices for every gid. That is, this property
        has length == len(gids) and row_indices[sc_gid] yields the row index of
        the target supply curve gid

        Returns
        -------
        ndarray
        """
        return self._sc_row_ind

    @property
    def col_indices(self):
        """Get a 1D array of col indices for every gid. That is, this property
        has length == len(gids) and col_indices[sc_gid] yields the col index of
        the target supply curve gid

        Returns
        -------
        ndarray
        """
        return self._sc_col_ind

    @property
    def points(self):
        """Get the summary dataframe of supply curve points.

        Returns
        -------
        _points : pd.DataFrame
            Supply curve points with columns for attributes of each sc point.
        """

        if self._points is None:
            self._points = pd.DataFrame({'row_ind': self.row_indices.copy(),
                                         'col_ind': self.col_indices.copy()})

            self._points.index.name = 'gid'  # sc_point_gid

        return self._points

    @staticmethod
    def _chunk_excl(arr, resolution):
        """Split an array into a list of arrays with len == resolution.

        Parameters
        ----------
        arr : np.ndarray
            1D array to be split into chunks.
        resolution : int
            Resolution of the chunks.

        Returns
        -------
        chunks : list
            List of arrays, each with length equal to self.resolution
            (except for the last array in the list which is the remainder).
        """

        chunks = get_chunk_ranges(len(arr), resolution)
        chunks = list(map(lambda i: np.arange(*i), chunks))

        return chunks

    @staticmethod
    def _excl_slices(arr, resolution):
        """Split row or col ind into slices of excl rows or slices

        Parameters
        ----------
        arr : np.ndarray
            1D array to be split into slices
        resolution : int
            Resolution of the sc points

        Returns
        -------
        slices : list
            List of arr slices, each with length equal to self.resolution
            (except for the last array in the list which is the remainder).
        """

        slices = get_chunk_ranges(len(arr), resolution)
        slices = list(map(lambda i: slice(*i), slices))

        return slices

    def get_sc_row_col_ind(self, gid):
        """Get the supply curve grid row and column index values corresponding
        to a supply curve gid.

        Parameters
        ----------
        gid : int
            Supply curve point gid.

        Returns
        -------
        row_ind : int
            Row index that the gid is located at in the sc grid.
        col_ind : int
            Column index that the gid is located at in the sc grid.
        """
        row_ind = self.points.loc[gid, 'row_ind']
        col_ind = self.points.loc[gid, 'col_ind']
        return row_ind, col_ind

    def get_excl_slices(self, gid):
        """Get the row and column slices of the exclusions grid corresponding
        to the supply curve point gid.

        Parameters
        ----------
        gid : int
            Supply curve point gid.

        Returns
        -------
        row_slice : slice
            Exclusions grid row slice corresponding to the sc point gid.
        col_slice : slice
            Exclusions grid col slice corresponding to the sc point gid.
        """

        if gid >= len(self):
            raise SupplyCurveError('Requested gid "{}" is out of bounds for '
                                   'supply curve points with length "{}".'
                                   .format(gid, len(self)))

        row_slice = self.excl_row_slices[self.row_indices[gid]]
        col_slice = self.excl_col_slices[self.col_indices[gid]]

        return row_slice, col_slice

    def get_flat_excl_ind(self, gid):
        """Get the index values of the flattened exclusions grid corresponding
        to the supply curve point gid.

        Parameters
        ----------
        gid : int
            Supply curve point gid.

        Returns
        -------
        excl_ind : np.ndarray
            Index values of the flattened exclusions grid corresponding to
            the SC gid.
        """

        row_slice, col_slice = self.get_excl_slices(gid)
        excl_ind = self.exclusions.iarr[row_slice, col_slice].flatten()

        return excl_ind

    def get_excl_points(self, dset, gid):
        """Get the exclusions data corresponding to a supply curve gid.

        Parameters
        ----------
        dset : str | int
            Used as the first arg in the exclusions __getitem__ slice.
            String can be "meta", integer can be layer number.
        gid : int
            Supply curve point gid.

        Returns
        -------
        excl_points : pd.DataFrame
            Exclusions data reduced to just the exclusion points associated
            with the requested supply curve gid.
        """

        row_slice, col_slice = self.get_excl_slices(gid)

        return self.exclusions[dset, row_slice, col_slice]

    def get_coord(self, gid):
        """Get the centroid coordinate for the supply curve gid point.

        Parameters
        ----------
        gid : int
            Supply curve point gid.

        Returns
        -------
        coord : tuple
            Two entry coordinate tuple: (latitude, longitude)
        """

        lat = self.latitude[gid]
        lon = self.longitude[gid]

        return (lat, lon)

    def valid_sc_points(self, tm_dset):
        """
        Determine which sc_point_gids contain resource gids and are thus
        valid supply curve points

        Parameters
        ----------
        tm_dset : str
            Techmap dataset name

        Returns
        -------
        valid_gids : ndarray
            Vector of valid sc_point_gids that contain resource gis
        """

        logger.info('Getting valid SC points from "{}"...'.format(tm_dset))

        valid_bool = np.zeros(self.n_rows * self.n_cols)
        tm = self._excls[tm_dset]

        gid = 0
        for r in self.excl_row_slices:
            for c in self.excl_col_slices:
                if np.any(tm[r, c] != -1):
                    valid_bool[gid] = 1
                gid += 1

        valid_gids = np.where(valid_bool == 1)[0].astype(np.uint32)

        logger.info('Found {} valid SC points out of {} total possible '
                    '(valid SC points that map to valid resource gids)'
                    .format(len(valid_gids), len(valid_bool)))

        return valid_gids

    def get_slice_lookup(self, sc_point_gids):
        """
        Get exclusion slices for all requested supply curve point gids

        Parameters
        ----------
        sc_point_gids : list | ndarray
            List or 1D array of sc_point_gids to get exclusion slices for

        Returns
        -------
        dict
            lookup mapping sc_point_gid to exclusion slice
        """
        return {g: self.get_excl_slices(g) for g in sc_point_gids}



# -*- coding: utf-8 -*-
"""
reV hybrids profile config

@author: ppinchuk
"""
import os
import glob
import logging

from reV.utilities import ModuleName
from reV.utilities.exceptions import PipelineError
from reV.config.base_analysis_config import AnalysisConfig

from rex.utilities.utilities import parse_year

logger = logging.getLogger(__name__)


class HybridsConfig(AnalysisConfig):
    """Hybrids config."""

    NAME = ModuleName.HYBRIDS
    REQUIREMENTS = ('solar_fpath', 'wind_fpath')

    def __init__(self, config):
        """
        Parameters
        ----------
        config : str | dict
            File path to config json (str), serialized json object (str),
            or dictionary with pre-extracted config.
        """
        super().__init__(config)
        self._default_ratio = 'solar_capacity/wind_capacity'

    @property
    def solar_fpath(self):
        """Get the solar data filepath"""

        fpath = self['solar_fpath']
        _raise_err_if_pipeline(fpath)

        return _glob_to_yearly_dict(fpath)

    @property
    def wind_fpath(self):
        """Get the wind data filepath"""

        fpath = self['wind_fpath']
        _raise_err_if_pipeline(fpath)

        return _glob_to_yearly_dict(fpath)

    @property
    def allow_solar_only(self):
        """Get the flag indicator for allowing solar-only capacity, """
        return bool(self.get('allow_solar_only', False))

    @property
    def allow_wind_only(self):
        """Get the flag indicator for allowing wind-only capacity, """
        return bool(self.get('allow_wind_only', False))

    @property
    def fillna(self):
        """Get the mapping for specifying input fill values. """
        return self.get('fillna', None)

    @property
    def limits(self):
        """Get the mapping for limiting hybrid meta values. """
        return self.get('limits', None)

    @property
    def ratio(self):
        """Get the ratio input string. """
        return self.get('ratio', self._default_ratio)

    @property
    def ratio_bounds(self):
        """Get the ratio_bounds input. """
        return self.get('ratio_bounds', None)


def _raise_err_if_pipeline(fpath):
    """Raise error if fpath input is 'PIPELINE'. """

    if fpath == 'PIPELINE':
        msg = ('Hybrids module cannot infer fpath from "PIPELINE" - '
               'input is ambiguous. Please specify both the solar and '
               'wind fpath before running hybrids module.')
        logger.error(msg)
        raise PipelineError(msg)


def _glob_to_yearly_dict(fpath):
    """Glob the filepaths into a dictionary based on years. """
    paths = {}
    for fp in glob.glob(fpath):
        fname = os.path.basename(fp)

        try:
            year = parse_year(fname)
        except RuntimeError:
            year = None

        paths.setdefault(year, []).append(fp)

    return paths



# -*- coding: utf-8 -*-
"""
reV representative profile config

Created on Mon Jan 28 11:43:27 2019

@author: gbuster
"""
import logging

from reV.utilities import ModuleName
from reV.utilities.exceptions import PipelineError
from reV.config.base_analysis_config import AnalysisConfig
from reV.pipeline.pipeline import Pipeline

logger = logging.getLogger(__name__)


class RepProfilesConfig(AnalysisConfig):
    """Representative Profiles config."""

    NAME = ModuleName.REP_PROFILES
    REQUIREMENTS = ('gen_fpath', 'rev_summary', 'reg_cols')

    def __init__(self, config):
        """
        Parameters
        ----------
        config : str | dict
            File path to config json (str), serialized json object (str),
            or dictionary with pre-extracted config.
        """
        super().__init__(config)

        self._default_cf_dset = 'cf_profile'
        self._default_rep_method = 'meanoid'
        self._default_err_method = 'rmse'
        self._default_weight = 'gid_counts'
        self._default_n_profiles = 1

    @property
    def gen_fpath(self):
        """Get the generation data filepath"""

        fpath = self['gen_fpath']

        if fpath == 'PIPELINE':
            targets = {ModuleName.MULTI_YEAR: 'fpath',
                       ModuleName.COLLECT: 'fpath',
                       ModuleName.GENERATION: 'fpath',
                       ModuleName.SUPPLY_CURVE_AGGREGATION: 'gen_fpath'}
            for target_module, target in targets.items():
                try:
                    fpath = Pipeline.parse_previous(
                        self.dirout, module=ModuleName.REP_PROFILES,
                        target=target, target_module=target_module)[0]
                except KeyError:
                    pass
                else:
                    break

            if fpath == 'PIPELINE':
                msg = 'Could not parse gen_fpath from previous pipeline jobs.'
                logger.error(msg)
                raise PipelineError(msg)
            else:
                logger.info('Rep profiles using the following '
                            'pipeline input for gen_fpath: {}'.format(fpath))

        return fpath

    @property
    def cf_dset(self):
        """Get the capacity factor dataset to get gen profiles from"""
        return self.get('cf_dset', self._default_cf_dset)

    @property
    def rev_summary(self):
        """Get the rev summary input arg."""

        fpath = self['rev_summary']

        if fpath == 'PIPELINE':
            target_modules = [
                ModuleName.SUPPLY_CURVE_AGGREGATION,
                ModuleName.SUPPLY_CURVE
            ]
            for target_module in target_modules:
                try:
                    fpath = Pipeline.parse_previous(
                        self.dirout, module=ModuleName.REP_PROFILES,
                        target='fpath', target_module=target_module)[0]
                except KeyError:
                    pass
                else:
                    break

            if fpath == 'PIPELINE':
                raise PipelineError('Could not parse rev_summary from '
                                    'previous pipeline jobs.')
            else:
                logger.info('Rep profiles using the following '
                            'pipeline input for rev_summary: {}'.format(fpath))

        return fpath

    @property
    def reg_cols(self):
        """Get the region columns input arg."""
        reg_cols = self.get('reg_cols', None)
        if isinstance(reg_cols, str):
            reg_cols = [reg_cols]

        return reg_cols

    @property
    def rep_method(self):
        """Get the representative profile method"""
        return self.get('rep_method', self._default_rep_method)

    @property
    def err_method(self):
        """Get the representative profile error method"""
        return self.get('err_method', self._default_err_method)

    @property
    def n_profiles(self):
        """Get the number of representative profiles to save."""
        return self.get('n_profiles', self._default_n_profiles)

    @property
    def weight(self):
        """Get the reV supply curve column to use for a weighted average in
        the representative profile meanoid algorithm."""
        return self.get('weight', self._default_weight)

    @property
    def aggregate_profiles(self):
        """Flag to calculate the aggregate (weighted meanoid) profile for each
        supply curve point. This behavior is instead of finding the single
        profile per region closest to the meanoid."""
        aggregate = bool(self.get('aggregate_profiles', False))
        if aggregate:
            self.check_overwrite_keys('aggregate_profiles', 'rep_method'
                                      'err_method', 'n_profiles')

        return aggregate



# -*- coding: utf-8 -*-
"""
reV utilities.
"""
from enum import Enum
import PySAM
from rex.utilities.loggers import log_versions as rex_log_versions
from reV.version import __version__


class ModuleName(str, Enum):
    """A collection of the module names available in reV.

    Each module name should match the name of the click command
    that will be used to invoke its respective cli. As of 3/1/2022,
    this means that all commands are lowercase with underscores
    replaced by dashes.

    Reference
    ---------
    See this line in the click source code to get the most up-to-date
    click name conversions:
        https://tinyurl.com/4rehbsvf

    """

    BESPOKE = 'bespoke'
    COLLECT = 'collect'
    ECON = 'econ'
    GENERATION = 'generation'
    HYBRIDS = 'hybrids'
    MULTI_YEAR = 'multi-year'
    NRWAL = 'nrwal'
    QA_QC = 'qa-qc'
    REP_PROFILES = 'rep-profiles'
    SUPPLY_CURVE = 'supply-curve'
    SUPPLY_CURVE_AGGREGATION = 'supply-curve-aggregation'

    @classmethod
    def all_names(cls):
        """All module names.

        Returns
        -------
        set
            The set of all module name strings.
        """
        # pylint: disable=no-member
        return {v.value for v in cls.__members__.values()}


def log_versions(logger):
    """Log package versions:
    - rex and reV to info
    - h5py, numpy, pandas, scipy, and PySAM to debug

    Parameters
    ----------
    logger : logging.Logger
        Logger object to log memory message to.
    """
    logger.info('Running with reV version {}'.format(__version__))
    rex_log_versions(logger)
    logger.debug('- PySAM version {}'.format(PySAM.__version__))



# -*- coding: utf-8 -*-
"""
Bespoke wind plant optimization CLI entry points.
"""

import click
import logging
import os
import pprint
import time
import re

from reV.bespoke.bespoke import BespokeWindPlants
from reV.config.bespoke import BespokeConfig
from reV.pipeline.status import Status
from reV.utilities.cli_dtypes import SAMFILES, SCPOINTS
from reV.utilities import ModuleName
from reV import __version__

from rex.utilities.cli_dtypes import (FLOAT, INT, STR, INTLIST, FLOATLIST,
                                      STRLIST, STR_OR_LIST, STRFLOAT)
from rex.utilities.hpc import SLURM
from rex.utilities.loggers import init_mult
from rex.utilities.utilities import get_class_properties, dict_str_load

logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('--name', '-n', default='reV-bespoke', type=STR,
              show_default=True,
              help='reV bespoke job name, by default "reV-bespoke".')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, name, verbose):
    """reV Bespoke Command Line Interface"""
    ctx.ensure_object(dict)
    ctx.obj['NAME'] = name
    ctx.obj['VERBOSE'] = verbose


@main.command()
@click.option('--config_file', '-c',
              required=True, type=click.Path(exists=True),
              help='reV bespoke configuration json file.')
@click.option('--points_range', '-pr',
              required=False, default=None, type=INTLIST,
              help='An optional input that specifies the (start, end) index '
              '(inclusive, exclusive) of the project points to analyze. If '
              'this is specified, the requested points are analyzed on a '
              'single worker. This input will override the config value.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def from_config(ctx, config_file, points_range, verbose):
    """Run reV gen from a config file."""

    config = (config_file if isinstance(config_file, BespokeConfig)
              else BespokeConfig(config_file))

    verbose = config.log_level == logging.DEBUG
    ctx.obj['VERBOSE'] = verbose = any([verbose, ctx.obj['VERBOSE']])

    if points_range is not None:
        config['points_range'] = points_range

    # take name from config
    name = ctx.obj['NAME'] = config.name

    # make output directory if does not exist
    if not os.path.exists(config.dirout):
        os.makedirs(config.dirout)

    # initialize loggers.
    node_tag = re.search('_node[0-9]*', name)
    if node_tag is None:
        init_mult(name, config.log_directory, modules=[__name__, 'reV', 'rex'],
                  verbose=verbose)

    # Initial log statements
    logger.info('Running reV Bespoke job with name "{}" from config file: "{}"'
                .format(name, config_file))
    logger.info('Target output directory: "{}"'.format(config.dirout))
    logger.info('Target logging directory: "{}"'.format(config.log_directory))
    logger.info('Source resource file: "{}"'.format(config.res_fpath))
    logger.info('Source exclusion file: "{}"'.format(config.excl_fpath))
    logger.info('Bespoke optimization objective function: "{}"'
                .format(config.objective_function))
    logger.info('Bespoke capital cost function: "{}"'
                .format(config.capital_cost_function))
    logger.info('Bespoke fixed operating cost function: "{}"'
                .format(config.fixed_operating_cost_function))
    logger.info('Bespoke variable operating cost function: "{}"'
                .format(config.variable_operating_cost_function))
    logger.info('Bespoke wake loss multiplier: "{}"'
                .format(config.wake_loss_multiplier))
    logger.info('The following project points were specified with '
                'points_range "{}": "{}"'
                .format(config.points_range, config.project_points))
    logger.info('The following SAM config files are available to this run:\n{}'
                .format(config.sam_files))
    logger.debug('The following exclusion dictionary was specified\n{}'
                 .format(pprint.pformat(config.excl_dict, indent=4)))
    logger.debug('The full configuration input is as follows:\n{}'
                 .format(pprint.pformat(config, indent=4)))

    # set config objects to be passed through invoke to direct methods
    ctx.obj['EXCL_FPATH'] = config.excl_fpath
    ctx.obj['RES_FPATH'] = config.res_fpath
    ctx.obj['TM_DSET'] = config.tm_dset
    ctx.obj['OBJECTIVE_FUNCTION'] = config.objective_function
    ctx.obj['CAPITAL_COST_FUNCTION'] = config.capital_cost_function
    ctx.obj['FIXED_OPERATING_COST_FUNCTION'] = \
        config.fixed_operating_cost_function
    ctx.obj['VARIABLE_OPERATING_COST_FUNCTION'] = \
        config.variable_operating_cost_function
    ctx.obj['POINTS'] = config.project_points
    ctx.obj['POINTS_RANGE'] = config.points_range
    ctx.obj['SAM_FILES'] = config.sam_files
    ctx.obj['MIN_SPACING'] = config.min_spacing
    ctx.obj['WAKE_LOSS_MULTIPLIER'] = config.wake_loss_multiplier
    ctx.obj['GA_KWARGS'] = config.ga_kwargs
    ctx.obj['OUTPUT_REQUEST'] = config.output_request
    ctx.obj['WS_BINS'] = config.ws_bins
    ctx.obj['WD_BINS'] = config.wd_bins
    ctx.obj['EXCL_DICT'] = config.excl_dict
    ctx.obj['AREA_FILTER_KERNEL'] = config.area_filter_kernel
    ctx.obj['MIN_AREA'] = config.min_area
    ctx.obj['RESOLUTION'] = config.resolution
    ctx.obj['EXCL_AREA'] = config.excl_area
    ctx.obj['DATA_LAYERS'] = config.data_layers
    ctx.obj['PRE_EXTRACT_INCLUSIONS'] = config.pre_extract_inclusions

    ctx.obj['LOG_DIR'] = config.log_directory
    ctx.obj['OUT_DIR'] = config.dirout
    ctx.obj['SITES_PER_WORKER'] = config.execution_control.sites_per_worker
    ctx.obj['MAX_WORKERS'] = config.execution_control.max_workers

    is_multi_node = (config.execution_control.option in ('eagle', 'slurm')
                     and config.points_range is None)
    nodes = config.execution_control.nodes if is_multi_node else 1

    pc = BespokeWindPlants._parse_points(config.excl_fpath,
                                         config.res_fpath,
                                         config.tm_dset,
                                         config.resolution,
                                         config.project_points,
                                         config.points_range,
                                         config.sam_files,
                                         sites_per_worker=1,
                                         workers=nodes)

    if len(pc) > 1:
        logger.info('Distributing the {} Bespoke project points to {} jobs.'
                    .format(len(pc.project_points), len(pc)))
        for i, pc_sub in enumerate(pc):
            logger.info('Creating distributed job submission for: {}'
                        .format(pc_sub.project_points))
            ctx.obj['NAME'] = name + '_node{}'.format(str(i).zfill(2))
            config._name = name + '_node{}'.format(str(i).zfill(2))
            ctx.invoke(from_config, config_file=config,
                       points_range=pc_sub.split_range, verbose=verbose)

    else:
        fout = name + '.h5'
        out_fpath = os.path.join(config.dirout, fout)
        ctx.obj['OUT_FPATH'] = out_fpath
        if config.execution_control.option == 'local':
            logger.info('Running Bespoke project with {} points '
                        'corresponding to points_range {}.'
                        .format(len(pc.project_points), config.points_range))
            status = Status.retrieve_job_status(config.dirout,
                                                ModuleName.BESPOKE,
                                                name)
            if status == 'successful':
                logger.info('Bespoke job with name "{}" was already '
                            'successfully run in directory: {}'
                            .format(name, config.dirout))
            else:
                job_attrs = {'hardware': 'local', 'fout': fout,
                             'dirout': config.dirout}
                Status.add_job(config.dirout, ModuleName.BESPOKE,
                               name, replace=True,
                               job_attrs=job_attrs)
                kws = {'excl_fpath': config.excl_fpath,
                       'res_fpath': config.res_fpath,
                       'out_fpath': out_fpath,
                       'tm_dset': config.tm_dset,
                       'objective_function':
                           config.objective_function,
                       'capital_cost_function':
                           config.capital_cost_function,
                       'fixed_operating_cost_function':
                           config.fixed_operating_cost_function,
                       'variable_operating_cost_function':
                           config.variable_operating_cost_function,
                       'points': config.project_points,
                       'sam_files': config.sam_files,
                       'points_range': config.points_range,
                       'min_spacing': config.min_spacing,
                       'wake_loss_multiplier': config.wake_loss_multiplier,
                       'ga_kwargs': config.ga_kwargs,
                       'output_request': config.output_request,
                       'ws_bins': config.ws_bins,
                       'wd_bins': config.wd_bins,
                       'excl_dict': config.excl_dict,
                       'area_filter_kernel': config.area_filter_kernel,
                       'min_area': config.min_area,
                       'resolution': config.resolution,
                       'excl_area': config.excl_area,
                       'log_dir': config.log_directory,
                       'max_workers': config.execution_control.max_workers,
                       'data_layers': config.data_layers,
                       'pre_extract_inclusions': config.pre_extract_inclusions,
                       'verbose': verbose}
                ctx.invoke(direct, **kws)

        elif config.execution_control.option in ('eagle', 'slurm'):
            stdout_path = os.path.join(config.log_directory, 'stdout')
            ctx.invoke(slurm,
                       alloc=config.execution_control.allocation,
                       walltime=config.execution_control.walltime,
                       feature=config.execution_control.feature,
                       memory=config.execution_control.memory,
                       module=config.execution_control.module,
                       conda_env=config.execution_control.conda_env,
                       stdout_path=stdout_path,
                       )


@main.command()
def valid_config_keys():
    """
    Echo the valid Bespoke config keys
    """
    click.echo(', '.join(get_class_properties(BespokeConfig)))


@main.group(invoke_without_command=True)
@click.option('--excl_fpath', '-exf', type=STR_OR_LIST, required=True,
              help='Single exclusions file (.h5) or a '
              'list of exclusion files (.h5, .h5).')
@click.option('--res_fpath', '-rf', type=STR, required=True,
              help='reV resource data file (e.g. WTK .h5 file). '
              'Can include a unix-style wildcard to source multiple years of '
              'data, e.g. res_fpath="/datasets/WIND/conus/v1.0.0/wtk_*.h5"')
@click.option('--out_fpath', '-of', type=STR, required=True,
              help='Filepath to save output data. Must be a .h5 path.')
@click.option('--tm_dset', '-tm', type=STR, required=True,
              help='Dataset in the exclusions file that maps the exclusions '
              'to the resource data file being analyzed.')
@click.option('--objective_function', '-obj', required=True, type=STR,
              help='The optimization objective function as a string, should'
              'return the objective to be minimized during optimization.'
              'Variables available are:'
              '- n_turbines: the number of turbines'
              '- system_capacity: wind plant capacity'
              '- aep: annual energy production'
              '- fixed_charge_rate: user input if included in sam sys config'
              '- capital_cost: the cap cost output by the cap cost function'
              '- fixed_operating_cost: the foc output by the foc function'
              '- variable_operating_cost: the voc output by the voc function'
              '- self.wind_plant: the SAM wind plant object, through which'
              'all SAM variables can be accessed')
@click.option('--capital_cost_function', '-cc',
              required=True, type=STR,
              help='The plant capital cost function as a string, must return '
              'the total capital cost in $. Has access to the same variables '
              'as the objective_function.')
@click.option('--fixed_operating_cost_function', '-foc',
              required=True, type=STR,
              help='The plant annual fixed operating cost function as a '
              'string, must return the fixed operating cost in $/year. Has '
              'access to the same variables as the objective_function.')
@click.option('--variable_operating_cost_function', '-voc',
              required=True, type=STR,
              help='The plant annual variable operating cost function as a '
              'string, must return the variable operating cost in $/kWh. Has '
              'access to the same variables as the objective_function.')
@click.option('--points', '-p',
              default=None, type=SCPOINTS, show_default=True,
              help='Project points to analyze. This can either be a string '
              'pointing to a csv with "gid" and "config" columns, or None '
              '(default) which will analyze all supply curve points in the '
              'given exclusion techmap.')
@click.option('--sam_files', '-sf', required=True, type=SAMFILES,
              help='SAM config files (required). Should be a single filepath '
              'to a sam config .json if points is None or a dictionary of '
              'key-value pairs where the keys map to the points "config" '
              'table column.')
@click.option('--points_range', '-pr',
              default=None, type=INTLIST, show_default=True,
              help='Optional list of (start, end) (inclusive, exclusive) '
              'index values that will slice the points input.')
@click.option('--min_spacing', '-ms',
              default='5x', type=STRFLOAT, show_default=True,
              help='Minimum spacing between turbines in meters. Can also be '
              'a string like "5x" (default) which is interpreted as 5 times '
              'the turbine rotor diameter.')
@click.option('--wake_loss_multiplier', '-wlm', type=FLOAT, default=1,
              show_default=True,
              help='A multiplier used to scale the annual energy lost due to '
              'wake losses. **IMPORTANT**: This multiplier will ONLY be '
              'applied during the optimization process and will NOT be '
              'come through in output values such as hourly generation '
              'profiles, aep, any of the cost functions, or even the output '
              'objective.')
@click.option('--ga_kwargs', '-gakw', type=STR, default=None,
              show_default=True,
              help='Dictionary of keyword arguments to pass to GA '
              'initialization.If `None`, default initialization values are '
              'used. See the reV.bespoke.gradient_free.GeneticAlgorithm class '
              'for a description of the allowed keyword arguments.')
@click.option('--output_request', '-or', type=STRLIST,
              default=['cf_mean', 'system_capacity'], show_default=True,
              help=('List of requested output variable names from SAM.'))
@click.option('--ws_bins', '-ws',
              default=[0, 20, 5], type=FLOATLIST, show_default=True,
              help='Get the windspeed binning arguments This should be a '
              '3-entry list with [start, stop, step] for the windspeed '
              'binning of the wind joint probability distribution. The stop '
              'value is inclusive, so ws_bins=[0, 20, 5] would result in '
              'four bins with bin edges [0, 5, 10, 15, 20]')
@click.option('--wd_bins', '-wd',
              default=[0, 360, 45], type=FLOATLIST, show_default=True,
              help='Get the winddirection binning arguments This should be a '
              '3-entry list with [start, stop, step] for the winddirection '
              'binning of the wind joint probability distribution. The stop '
              'value is inclusive, so ws_bins=[0, 360, 90] would result in '
              'four bins with bin edges [0, 90, 180, 270, 360]')
@click.option('--excl_dict', '-exd', type=STR, default=None,
              show_default=True,
              help='String representation of a dictionary of exclusion '
              'LayerMask arguments {layer: {kwarg: value}} where layer '
              'is a dataset in excl_fpath and kwarg can be '
              '"exclude_values", "exclude_range", "include_values", '
              '"include_range", "include_weights", "force_include_values", '
              '"force_include_range", "use_as_weights", "weight", '
              '"exclude_nodata", and/or "nodata_value".')
@click.option('--area_filter_kernel', '-afk', type=STR, default='queen',
              show_default=True,
              help='Contiguous area filter kernel name ("queen", "rook").')
@click.option('--min_area', '-ma', type=FLOAT, default=None,
              show_default=True,
              help='Contiguous area filter minimum area, default is None '
              '(No minimum area filter).')
@click.option('--resolution', '-r', type=INT, default=64,
              show_default=True,
              help='Number of exclusion points along a squares edge to '
              'include in an aggregated supply curve point.')
@click.option('--excl_area', '-ea',
              type=FLOAT, default=None, show_default=True,
              help='Area of an exclusion pixel in km2. None will try to '
              'infer the area from the profile transform attribute in '
              'excl_fpath.')
@click.option('--log_dir', '-lo', default='./out/log_bespoke', type=STR,
              help='Bespoke log file directory. Default is ./out/log_bespoke')
@click.option('--max_workers', '-mw', type=INT, default=None,
              show_default=True,
              help='Number of workers. Use 1 for serial, None for all cores.')
@click.option('--data_layers', '-d', type=STR, default=None,
              show_default=True,
              help='String representation of a dictionary of additional data '
              'layers to include in the aggregation e.g. '
              '{"slope": {"dset": "srtm_slope", "method": "mean", '
              '"fpath": "./excl.h5"}}')
@click.option('-pre', '--pre_extract_inclusions', is_flag=True,
              help='Optional flag to pre-extract/compute the inclusion mask '
              'from the provided excl_dict, by default False. Typically '
              'faster to compute the inclusion mask on the fly with parallel '
              'workers.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def direct(ctx, excl_fpath, res_fpath, out_fpath, tm_dset, objective_function,
           capital_cost_function, fixed_operating_cost_function,
           variable_operating_cost_function, points, sam_files, points_range,
           min_spacing, wake_loss_multiplier, ga_kwargs, output_request,
           ws_bins, wd_bins, excl_dict, area_filter_kernel, min_area,
           resolution, excl_area, log_dir, max_workers, data_layers,
           pre_extract_inclusions, verbose):
    """Run reV Bespoke directly w/o a config file."""
    ctx.obj['EXCL_FPATH'] = excl_fpath
    ctx.obj['RES_FPATH'] = res_fpath
    ctx.obj['OUT_FPATH'] = out_fpath
    ctx.obj['TM_DSET'] = tm_dset
    ctx.obj['OBJECTIVE_FUNCTION'] = objective_function
    ctx.obj['CAPITAL_COST_FUNCTION'] = capital_cost_function
    ctx.obj['FIXED_OPERATING_COST_FUNCTION'] = fixed_operating_cost_function
    ctx.obj['VARIABLE_OPERATING_COST_FUNCTION'] = \
        variable_operating_cost_function
    ctx.obj['POINTS'] = points
    ctx.obj['POINTS_RANGE'] = points_range
    ctx.obj['SAM_FILES'] = sam_files
    ctx.obj['MIN_SPACING'] = min_spacing
    ctx.obj['WAKE_LOSS_MULTIPLIER'] = wake_loss_multiplier
    ctx.obj['GA_KWARGS'] = ga_kwargs
    ctx.obj['OUTPUT_REQUEST'] = output_request
    ctx.obj['WS_BINS'] = ws_bins
    ctx.obj['WD_BINS'] = wd_bins
    ctx.obj['EXCL_DICT'] = excl_dict
    ctx.obj['AREA_FILTER_KERNEL'] = area_filter_kernel
    ctx.obj['MIN_AREA'] = min_area
    ctx.obj['RESOLUTION'] = resolution
    ctx.obj['EXCL_AREA'] = excl_area
    ctx.obj['DATA_LAYERS'] = data_layers
    ctx.obj['PRE_EXTRACT_INCLUSIONS'] = pre_extract_inclusions
    ctx.obj['LOG_DIR'] = log_dir
    verbose = any([verbose, ctx.obj['VERBOSE']])

    name = ctx.obj['NAME']

    if ctx.invoked_subcommand is None:
        init_mult(name, log_dir, modules=[__name__, 'reV', 'rex'],
                  verbose=verbose, node=True)

        for key, val in ctx.obj.items():
            logger.debug('ctx var passed to local method: "{}" : "{}" with '
                         'type "{}"'.format(key, val, type(val)))

        logger.info('Bespoke local is being run with with job name "{}" and '
                    'resource file: {}. Target output path is: {}'
                    .format(name, res_fpath, out_fpath))

        if isinstance(data_layers, str):
            data_layers = dict_str_load(data_layers)

        if isinstance(excl_dict, str):
            excl_dict = dict_str_load(excl_dict)

        if isinstance(ga_kwargs, str):
            ga_kwargs = dict_str_load(ga_kwargs)

        t0 = time.time()

        try:
            # Execute the Bespoke module with smart data flushing.
            BespokeWindPlants.run(
                excl_fpath, res_fpath, tm_dset,
                objective_function,
                capital_cost_function,
                fixed_operating_cost_function,
                variable_operating_cost_function,
                points, sam_files,
                points_range=points_range,
                min_spacing=min_spacing,
                wake_loss_multiplier=wake_loss_multiplier,
                ga_kwargs=ga_kwargs,
                output_request=output_request,
                ws_bins=ws_bins,
                wd_bins=wd_bins,
                excl_dict=excl_dict,
                area_filter_kernel=area_filter_kernel,
                min_area=min_area,
                resolution=resolution,
                excl_area=excl_area,
                data_layers=data_layers,
                pre_extract_inclusions=pre_extract_inclusions,
                max_workers=max_workers,
                out_fpath=out_fpath)
        except Exception as e:
            msg = ('reV Bespoke optimization failed. Received the '
                   'following error:\n{}'.format(e))
            logger.exception(msg)
            raise RuntimeError(msg) from e

        tmp_str = ' with points range {}'.format(points_range)
        out_dir, fout = os.path.split(out_fpath)
        runtime = (time.time() - t0) / 60
        logger.info('Bespoke compute complete for project points "{0}"{1}. '
                    'Time elapsed: {2:.2f} min. Target output dir: {3}'
                    .format(points, tmp_str if points_range else '',
                            runtime, out_dir))

        # add job to reV status file.
        status = {'dirout': out_dir, 'fout': fout, 'job_status': 'successful',
                  'runtime': runtime, 'res_fpath': res_fpath,
                  'excl_fpath': excl_fpath, 'tm_dset': tm_dset,
                  'objective_function': objective_function,
                  'capital_cost_function': capital_cost_function,
                  'fixed_operating_cost_function':
                      fixed_operating_cost_function,
                  'variable_operating_cost_function':
                      variable_operating_cost_function,
                  'excl_dict': excl_dict}
        Status.make_job_file(out_dir, ModuleName.BESPOKE, name, status)


def get_node_cmd(name, kwargs):
    """Make a CLI command string."""
    arg_main = '-n {}'.format(SLURM.s(name))

    arg_direct = [
        '-exf {}'.format(SLURM.s(kwargs['EXCL_FPATH'])),
        '-rf {}'.format(SLURM.s(kwargs['RES_FPATH'])),
        '-of {}'.format(SLURM.s(kwargs['OUT_FPATH'])),
        '-tm {}'.format(SLURM.s(kwargs['TM_DSET'])),
        '-obj {}'.format(SLURM.s(kwargs['OBJECTIVE_FUNCTION'])),
        '-cc {}'.format(SLURM.s(kwargs['CAPITAL_COST_FUNCTION'])),
        '-foc {}'.format(SLURM.s(kwargs['FIXED_OPERATING_COST_FUNCTION'])),
        '-voc {}'.format(SLURM.s(kwargs['VARIABLE_OPERATING_COST_FUNCTION'])),
        '-p {}'.format(SLURM.s(kwargs['POINTS'])),
        '-sf {}'.format(SLURM.s(kwargs['SAM_FILES'])),
        '-pr {}'.format(SLURM.s(kwargs['POINTS_RANGE'])),
        '-ms {}'.format(SLURM.s(kwargs['MIN_SPACING'])),
        '-wlm {}'.format(SLURM.s(kwargs['WAKE_LOSS_MULTIPLIER'])),
        '-gakw {}'.format(SLURM.s(kwargs['GA_KWARGS'])),
        '-or {}'.format(SLURM.s(kwargs['OUTPUT_REQUEST'])),
        '-ws {}'.format(SLURM.s(kwargs['WS_BINS'])),
        '-wd {}'.format(SLURM.s(kwargs['WD_BINS'])),
        '-exd {}'.format(SLURM.s(kwargs['EXCL_DICT'])),
        '-afk {}'.format(SLURM.s(kwargs['AREA_FILTER_KERNEL'])),
        '-ma {}'.format(SLURM.s(kwargs['MIN_AREA'])),
        '-r {}'.format(SLURM.s(kwargs['RESOLUTION'])),
        '-ea {}'.format(SLURM.s(kwargs['EXCL_AREA'])),
        '-d {}'.format(SLURM.s(kwargs['DATA_LAYERS'])),
        '-lo {}'.format(SLURM.s(kwargs['LOG_DIR'])),
        '-mw {}'.format(SLURM.s(kwargs['MAX_WORKERS']))]

    if kwargs['PRE_EXTRACT_INCLUSIONS']:
        arg_direct.append('-pre')

    if kwargs['VERBOSE']:
        arg_direct.append('-v')

    cmd = ('python -m reV.bespoke.cli_bespoke '
           '{arg_main} direct {arg_direct}'
           .format(arg_main=arg_main,
                   arg_direct=' '.join(arg_direct)))
    logger.debug('Creating the following command line call:\n\t{}'.format(cmd))

    return cmd


@direct.command()
@click.option('--alloc', '-a', required=True, type=STR,
              help='SLURM allocation account name.')
@click.option('--walltime', '-wt', default=1.0, type=float,
              show_default=True,
              help='SLURM walltime request in hours. Default is 1.0')
@click.option('--feature', '-l', default=None, type=STR,
              show_default=True,
              help=('Additional flags for SLURM job. Format is "--qos=high" '
                    'or "--depend=[state:job_id]". Default is None.'))
@click.option('--memory', '-mem', default=None, type=INT,
              show_default=True,
              help='SLURM node memory request in GB. Default is None')
@click.option('--module', '-mod', default=None, type=STR,
              show_default=True,
              help='Module to load')
@click.option('--conda_env', '-env', default=None, type=STR,
              show_default=True,
              help='Conda env to activate')
@click.option('--stdout_path', '-sout', default=None, type=STR,
              show_default=True,
              help='Subprocess standard output path. Default is in out_dir.')
@click.pass_context
def slurm(ctx, alloc, walltime, feature, memory, module, conda_env,
          stdout_path):
    """slurm (Eagle) submission tool for reV Bespoke wind plant optimization"""
    name = ctx.obj['NAME']
    log_dir = ctx.obj['LOG_DIR']
    out_fpath = ctx.obj['OUT_FPATH']
    out_dir = os.path.dirname(out_fpath)

    if stdout_path is None:
        stdout_path = os.path.join(log_dir, 'stdout/')

    cmd = get_node_cmd(name, ctx.obj)

    slurm_manager = ctx.obj.get('SLURM_MANAGER', None)
    if slurm_manager is None:
        slurm_manager = SLURM()
        ctx.obj['SLURM_MANAGER'] = slurm_manager

    status = Status.retrieve_job_status(out_dir, ModuleName.BESPOKE,
                                        name, hardware='slurm',
                                        subprocess_manager=slurm_manager)

    msg = ''
    if status == 'successful':
        msg = ('Job "{}" is successful in status json found in "{}", '
               'not re-running.'
               .format(name, out_dir))
    elif 'fail' not in str(status).lower() and status is not None:
        msg = ('Job "{}" was found with status "{}", not resubmitting'
               .format(name, status))
    else:
        logger.info('Running reV Bespoke on SLURM with '
                    'node name "{}"'.format(name))
        out = slurm_manager.sbatch(cmd, alloc=alloc, memory=memory,
                                   walltime=walltime, feature=feature,
                                   name=name, stdout_path=stdout_path,
                                   conda_env=conda_env, module=module)[0]
        if out:
            msg = ('Kicked off reV Bespoke job "{}" (SLURM jobid #{}).'
                   .format(name, out))
            Status.add_job(
                out_dir, ModuleName.BESPOKE, name, replace=True,
                job_attrs={'job_id': out, 'hardware': 'slurm',
                           'fout': '{}.csv'.format(name), 'dirout': out_dir})

    click.echo(msg)
    logger.info(msg)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running reV Bespoke CLI')
        raise
